import unittest
from unittest.mock import MagicMock, call
from pathlib import Path

from subt.artf_reporter import ArtifactReporter


class ArtifactReporterTest(unittest.TestCase):

    def test_artf_repeat(self):
        # default behavior
        config = {}
        bus = MagicMock()
        reporter = ArtifactReporter(config, bus)
        bus.listen = MagicMock(return_value=(1, 'artf_xyz', [['BACKPACK', [100, 200, -3], "robot-name", None]]))
        reporter.update()
        bus.publish.assert_called()
        bus.reset_mock()
        for sim_time_sec in range(20):
            bus.listen = MagicMock(return_value=(1, 'sim_time_sec', sim_time_sec))
            reporter.update()
        bus.publish.assert_not_called()

        # now with repetitions
        config = {
            'repeat_report_sec': 10
        }
        bus = MagicMock()
        reporter = ArtifactReporter(config, bus)
        bus.listen = MagicMock(return_value=(1, 'artf_xyz', [['BACKPACK', [100, 200, -3], "name", None]]))
        reporter.update()
        bus.publish.assert_called()
        bus.reset_mock()
        for sim_time_sec in range(20):
            bus.listen = MagicMock(return_value=(1, 'sim_time_sec', sim_time_sec))
            reporter.update()
        # well, I am not sure how to ensure only 2 calls :(
        bus.publish.assert_has_calls([call('artf_cmd', b'artf BACKPACK 0.10 0.20 -0.00\n'),
                                      call('artf_all', [['BACKPACK', [100, 200, -3], "name", None]]),]*2)

    def test_artf_empty_repeat(self):
        config = {
            'repeat_report_sec': 10
        }
        bus = MagicMock()
        reporter = ArtifactReporter(config, bus)
        sim_time_sec = 0
        bus.listen = MagicMock(return_value=(1, 'sim_time_sec', sim_time_sec))
        reporter.update()
        bus.publish.assert_not_called()

    def test_artf_report_aggregation(self):
        config = {
            'repeat_report_sec': 10
        }
        bus = MagicMock()
        reporter = ArtifactReporter(config, bus)
        bus.listen = MagicMock(return_value=(1, 'artf_xyz', [['TYPE_BACKPACK', [100, 200, -3], "RoboX", None]]))
        reporter.update()
        bus.listen = MagicMock(return_value=(1, 'artf_xyz', [['TYPE_ROPE', [10, 20, 30], "RobotX", None]]))
        reporter.update()
        bus.reset_mock()
        sim_time_sec = 0
        bus.listen = MagicMock(return_value=(1, 'sim_time_sec', sim_time_sec))
        reporter.update()
        bus.publish.assert_has_calls([call('artf_cmd', b'artf TYPE_BACKPACK 0.10 0.20 -0.00\n'),
                                      call('artf_cmd', b'artf TYPE_ROPE 0.01 0.02 0.03\n')])

        # keep list unique
        bus.listen = MagicMock(return_value=(1, 'artf_xyz', [['TYPE_BACKPACK', 100, 200, -3]]))
        reporter.update()
        bus.reset_mock()
        sim_time_sec = 0
        bus.listen = MagicMock(return_value=(1, 'sim_time_sec', sim_time_sec))
        reporter.update()
        # the same issue, that I do not want subset, but 1:1 correspondence - TODO more strict assert
        bus.publish.assert_has_calls([call('artf_cmd', b'artf TYPE_BACKPACK 0.10 0.20 -0.00\n'),
                                      call('artf_cmd', b'artf TYPE_ROPE 0.01 0.02 0.03\n')])

    def test_on_base_station(self):
        data = {'report_id': 1, 'artifact_type': 0, 'artifact_position': [69.84, -4.86, 0.4],
                'report_status': 'scored', 'score_change': 1}
        bus = MagicMock()
        reporter = ArtifactReporter(config={}, bus=bus)
        reporter.on_base_station(data)
        self.assertEqual(reporter.artf_xyz_accumulated, [])
        bus.publish.assert_not_called()  # nothing to report/share
        reporter.artf_xyz_accumulated = [['TYPE_BACKPACK', [69845, -4860, 396], 'A150L', None]]
        reporter.on_base_station(data)
        self.assertEqual(reporter.artf_xyz_accumulated, [['TYPE_BACKPACK', [69845, -4860, 396], 'A150L', True]])
        bus.publish.assert_called()  # make sure that new confirmed artifact is propagated

        bus.reset_mock()
        reporter.on_base_station(data)
        self.assertEqual(reporter.artf_xyz_accumulated, [['TYPE_BACKPACK', [69845, -4860, 396], 'A150L', True]])
        bus.publish.assert_not_called()  # the information did not change - no hurry to spread it

        data['score_change'] = 0  # wrongly reported artifact
        reporter.on_base_station(data)
        self.assertEqual(reporter.artf_xyz_accumulated, [['TYPE_BACKPACK', [69845, -4860, 396], 'A150L', False]])

    def test_on_artf_xyz(self):
        # update from detector or from radio
        bus = MagicMock()
        reporter = ArtifactReporter(config={}, bus=bus)
        reporter.artf_xyz_accumulated = [['TYPE_BACKPACK', [69758, -5143, 533], 'A150L', True], ['TYPE_BACKPACK', [70044, -5289, 438], 'B10W150L', False]]
        reporter.on_artf_xyz([['TYPE_BACKPACK', [69758, -5143, 533], 'A150L', None]])
        self.assertEqual(reporter.artf_xyz_accumulated, [['TYPE_BACKPACK', [69758, -5143, 533], 'A150L', True],
                                         ['TYPE_BACKPACK', [70044, -5289, 438], 'B10W150L', False]])

    def test_duplicity_reports(self):
        bus = MagicMock()
        reporter = ArtifactReporter(config={}, bus=bus)
        self.assertEqual(len(reporter.artf_xyz_accumulated), 0)
        reporter.on_artf_xyz([['TYPE_BACKPACK', [69758, -5143, 533], 'A150L', True]])
        reporter.on_artf_xyz([['TYPE_BACKPACK', [69758 + 1000, -5143, 533], 'A150L', None]])  # 1m offset
        self.assertEqual(len(reporter.artf_xyz_accumulated), 1)
        bus.publish.assert_not_called()

        # backpack is confirmed and phone is tested as different artifact
        reporter.on_artf_xyz([['TYPE_PHONE', [69758 + 2000, -5143, 533], 'A150L', None]])  # 2m offset
        self.assertEqual(len(reporter.artf_xyz_accumulated), 2)


# vim: expandtab sw=4 ts=4




# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ign.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ign.proto',
  package='',
  syntax='proto3',
  serialized_pb=_b('\n\tign.proto\"!\n\x04Time\x12\x0b\n\x03sec\x18\x01 \x01(\x03\x12\x0c\n\x04nsec\x18\x02 \x01(\x05\"\\\n\x06Header\x12\x14\n\x05stamp\x18\x01 \x01(\x0b\x32\x05.Time\x12\x19\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32\x0b.Header.Map\x1a!\n\x03Map\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x03(\t\"D\n\x08Vector3d\x12\x17\n\x06header\x18\x01 \x01(\x0b\x32\x07.Header\x12\t\n\x01x\x18\x02 \x01(\x01\x12\t\n\x01y\x18\x03 \x01(\x01\x12\t\n\x01z\x18\x04 \x01(\x01\"Q\n\nQuaternion\x12\x17\n\x06header\x18\x01 \x01(\x0b\x32\x07.Header\x12\t\n\x01x\x18\x02 \x01(\x01\x12\t\n\x01y\x18\x03 \x01(\x01\x12\t\n\x01z\x18\x04 \x01(\x01\x12\t\n\x01w\x18\x05 \x01(\x01\"x\n\x04Pose\x12\x17\n\x06header\x18\x01 \x01(\x0b\x32\x07.Header\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\r\x12\x1b\n\x08position\x18\x04 \x01(\x0b\x32\t.Vector3d\x12 \n\x0borientation\x18\x05 \x01(\x0b\x32\x0b.Quaternion\"6\n\x06Pose_V\x12\x17\n\x06header\x18\x01 \x01(\x0b\x32\x07.Header\x12\x13\n\x04pose\x18\x02 \x03(\x0b\x32\x05.Pose\"-\n\x08\x41rtifact\x12\x0c\n\x04type\x18\x01 \x01(\r\x12\x13\n\x04pose\x18\x02 \x01(\x0b\x32\x05.Pose\"\xb5\x01\n\rArtifactScore\x12\x11\n\treport_id\x18\x01 \x01(\r\x12\x1b\n\x08\x61rtifact\x18\x02 \x01(\x0b\x32\t.Artifact\x12!\n\x12submitted_datetime\x18\x03 \x01(\x0b\x32\x05.Time\x12\x17\n\x08sim_time\x18\x04 \x01(\x0b\x32\x05.Time\x12\x0b\n\x03run\x18\x05 \x01(\r\x12\x15\n\rreport_status\x18\x06 \x01(\t\x12\x14\n\x0cscore_change\x18\x07 \x01(\x05\"2\n\tStringMsg\x12\x17\n\x06header\x18\x01 \x01(\x0b\x32\x07.Header\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\tb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TIME = _descriptor.Descriptor(
  name='Time',
  full_name='Time',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sec', full_name='Time.sec', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='nsec', full_name='Time.nsec', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13,
  serialized_end=46,
)


_HEADER_MAP = _descriptor.Descriptor(
  name='Map',
  full_name='Header.Map',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Header.Map.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='Header.Map.value', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=140,
)

_HEADER = _descriptor.Descriptor(
  name='Header',
  full_name='Header',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='stamp', full_name='Header.stamp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='data', full_name='Header.data', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_HEADER_MAP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=48,
  serialized_end=140,
)


_VECTOR3D = _descriptor.Descriptor(
  name='Vector3d',
  full_name='Vector3d',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='Vector3d.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='x', full_name='Vector3d.x', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y', full_name='Vector3d.y', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='z', full_name='Vector3d.z', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=142,
  serialized_end=210,
)


_QUATERNION = _descriptor.Descriptor(
  name='Quaternion',
  full_name='Quaternion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='Quaternion.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='x', full_name='Quaternion.x', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y', full_name='Quaternion.y', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='z', full_name='Quaternion.z', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='w', full_name='Quaternion.w', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=212,
  serialized_end=293,
)


_POSE = _descriptor.Descriptor(
  name='Pose',
  full_name='Pose',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='Pose.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='Pose.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='id', full_name='Pose.id', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='position', full_name='Pose.position', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='orientation', full_name='Pose.orientation', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=295,
  serialized_end=415,
)


_POSE_V = _descriptor.Descriptor(
  name='Pose_V',
  full_name='Pose_V',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='Pose_V.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pose', full_name='Pose_V.pose', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=417,
  serialized_end=471,
)


_ARTIFACT = _descriptor.Descriptor(
  name='Artifact',
  full_name='Artifact',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='Artifact.type', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pose', full_name='Artifact.pose', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=473,
  serialized_end=518,
)


_ARTIFACTSCORE = _descriptor.Descriptor(
  name='ArtifactScore',
  full_name='ArtifactScore',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='report_id', full_name='ArtifactScore.report_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='artifact', full_name='ArtifactScore.artifact', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='submitted_datetime', full_name='ArtifactScore.submitted_datetime', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sim_time', full_name='ArtifactScore.sim_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='run', full_name='ArtifactScore.run', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='report_status', full_name='ArtifactScore.report_status', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='score_change', full_name='ArtifactScore.score_change', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=521,
  serialized_end=702,
)


_STRINGMSG = _descriptor.Descriptor(
  name='StringMsg',
  full_name='StringMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='StringMsg.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='data', full_name='StringMsg.data', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=704,
  serialized_end=754,
)

_HEADER_MAP.containing_type = _HEADER
_HEADER.fields_by_name['stamp'].message_type = _TIME
_HEADER.fields_by_name['data'].message_type = _HEADER_MAP
_VECTOR3D.fields_by_name['header'].message_type = _HEADER
_QUATERNION.fields_by_name['header'].message_type = _HEADER
_POSE.fields_by_name['header'].message_type = _HEADER
_POSE.fields_by_name['position'].message_type = _VECTOR3D
_POSE.fields_by_name['orientation'].message_type = _QUATERNION
_POSE_V.fields_by_name['header'].message_type = _HEADER
_POSE_V.fields_by_name['pose'].message_type = _POSE
_ARTIFACT.fields_by_name['pose'].message_type = _POSE
_ARTIFACTSCORE.fields_by_name['artifact'].message_type = _ARTIFACT
_ARTIFACTSCORE.fields_by_name['submitted_datetime'].message_type = _TIME
_ARTIFACTSCORE.fields_by_name['sim_time'].message_type = _TIME
_STRINGMSG.fields_by_name['header'].message_type = _HEADER
DESCRIPTOR.message_types_by_name['Time'] = _TIME
DESCRIPTOR.message_types_by_name['Header'] = _HEADER
DESCRIPTOR.message_types_by_name['Vector3d'] = _VECTOR3D
DESCRIPTOR.message_types_by_name['Quaternion'] = _QUATERNION
DESCRIPTOR.message_types_by_name['Pose'] = _POSE
DESCRIPTOR.message_types_by_name['Pose_V'] = _POSE_V
DESCRIPTOR.message_types_by_name['Artifact'] = _ARTIFACT
DESCRIPTOR.message_types_by_name['ArtifactScore'] = _ARTIFACTSCORE
DESCRIPTOR.message_types_by_name['StringMsg'] = _STRINGMSG

Time = _reflection.GeneratedProtocolMessageType('Time', (_message.Message,), dict(
  DESCRIPTOR = _TIME,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:Time)
  ))
_sym_db.RegisterMessage(Time)

Header = _reflection.GeneratedProtocolMessageType('Header', (_message.Message,), dict(

  Map = _reflection.GeneratedProtocolMessageType('Map', (_message.Message,), dict(
    DESCRIPTOR = _HEADER_MAP,
    __module__ = 'ign_pb2'
    # @@protoc_insertion_point(class_scope:Header.Map)
    ))
  ,
  DESCRIPTOR = _HEADER,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:Header)
  ))
_sym_db.RegisterMessage(Header)
_sym_db.RegisterMessage(Header.Map)

Vector3d = _reflection.GeneratedProtocolMessageType('Vector3d', (_message.Message,), dict(
  DESCRIPTOR = _VECTOR3D,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:Vector3d)
  ))
_sym_db.RegisterMessage(Vector3d)

Quaternion = _reflection.GeneratedProtocolMessageType('Quaternion', (_message.Message,), dict(
  DESCRIPTOR = _QUATERNION,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:Quaternion)
  ))
_sym_db.RegisterMessage(Quaternion)

Pose = _reflection.GeneratedProtocolMessageType('Pose', (_message.Message,), dict(
  DESCRIPTOR = _POSE,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:Pose)
  ))
_sym_db.RegisterMessage(Pose)

Pose_V = _reflection.GeneratedProtocolMessageType('Pose_V', (_message.Message,), dict(
  DESCRIPTOR = _POSE_V,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:Pose_V)
  ))
_sym_db.RegisterMessage(Pose_V)

Artifact = _reflection.GeneratedProtocolMessageType('Artifact', (_message.Message,), dict(
  DESCRIPTOR = _ARTIFACT,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:Artifact)
  ))
_sym_db.RegisterMessage(Artifact)

ArtifactScore = _reflection.GeneratedProtocolMessageType('ArtifactScore', (_message.Message,), dict(
  DESCRIPTOR = _ARTIFACTSCORE,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:ArtifactScore)
  ))
_sym_db.RegisterMessage(ArtifactScore)

StringMsg = _reflection.GeneratedProtocolMessageType('StringMsg', (_message.Message,), dict(
  DESCRIPTOR = _STRINGMSG,
  __module__ = 'ign_pb2'
  # @@protoc_insertion_point(class_scope:StringMsg)
  ))
_sym_db.RegisterMessage(StringMsg)


# @@protoc_insertion_point(module_scope)



"""
  Applications launcher

example:
    launch(app=GoOneMeter, description='Go One Meter', prefix='go1m-')
"""
import argparse

from osgar.record import record
from osgar.replay import replay
from osgar.lib.config import config_load


def launch(app, description, prefix):
    parser = argparse.ArgumentParser(description=description)
    subparsers = parser.add_subparsers(help='sub-command help', dest='command')
    subparsers.required = True
    parser_run = subparsers.add_parser('run', help='run on real HW')
    parser_run.add_argument('config', nargs='+', help='configuration file')
    parser_run.add_argument('--note', help='add description')

    parser_replay = subparsers.add_parser('replay', help='replay from logfile')
    parser_replay.add_argument('logfile', help='recorded log file')
    parser_replay.add_argument('--force', '-F', dest='force', action='store_true', help='force replay even for failing output asserts')
    parser_replay.add_argument('--config', nargs='+', help='force alternative configuration file')
    parser_replay.add_argument('--verbose', '-v', help="verbose mode", action='store_true')
    args = parser.parse_args()

    if args.command == 'replay':
        args.module = 'app'
        game = replay(args, application=app)
        game.verbose = args.verbose
        game.run()

    elif args.command == 'run':
        cfg = config_load(*args.config, application=app)
        record(cfg, log_prefix=prefix)

# vim: expandtab sw=4 ts=4



import contextlib

from threading import Thread

import zmq

from osgar.bus import BusShutdownException
import osgar.lib.serialize

class Pull:

    def __init__(self, config, bus):
        bus.register(*config['outputs'])
        self.endpoint = config.get('endpoint', 'tcp://127.0.0.1:5565')
        self.timeout = config.get('timeout', 1) # default recv timeout 1s
        self.thread = Thread(target=self.run)
        self.thread.name = bus.name
        self.bus = bus

    def start(self):
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout=timeout)

    def run(self):
        context = zmq.Context.instance()
        socket = context.socket(zmq.PULL)
        # https://stackoverflow.com/questions/7538988/zeromq-how-to-prevent-infinite-wait
        socket.RCVTIMEO = int(self.timeout * 1000)  # convert to milliseconds
        socket.connect(self.endpoint)

        with contextlib.closing(socket):
            while self.bus.is_alive():
                try:
                    channel, raw = socket.recv_multipart()
                    message = osgar.lib.serialize.deserialize(raw)
                    self.bus.publish(channel.decode('ascii'), message)
                except zmq.error.Again:
                    pass

    def request_stop(self):
        self.bus.shutdown()





"""
  Log video stream provided by OpenCV camera
"""

import cv2
from threading import Thread

class LogOpenCVCamera:
    def __init__(self, config, bus):
        self.input_thread = Thread(target=self.run_input, daemon=True)
        self.bus = bus
        bus.register('raw')

        port = config.get('port', 0)
        self.cap = cv2.VideoCapture(port)
        self.sleep = config.get('sleep')

    def start(self):
        self.input_thread.start()

    def join(self, timeout=None):
        self.input_thread.join(timeout=timeout)

    def run_input(self):
        while self.bus.is_alive():
            # Capture frame-by-frame
            ret, frame = self.cap.read()
            if ret:
                retval, data = cv2.imencode('*.jpeg', frame)
                if len(data) > 0:
                    self.bus.publish('raw', data.tobytes())
                if self.sleep is not None:
                    self.bus.sleep(self.sleep)
        self.cap.release()

    def request_stop(self):
        self.bus.shutdown()

# vim: expandtab sw=4 ts=4



"""
  OSGAR Teambase for Virtual
"""
from ast import literal_eval

from osgar.node import Node


class Teambase(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register()

        self.robot_name = config.get('robot_name')
        self.start_time = None  # unknown
        self.finish_time = None  # infinite
        if self.robot_name is not None:
            self.finish_time = int(self.robot_name[1:])  # T100 is accepted for example
        self.robot_positions = {}
        self.debug_arr = []
        self.artifacts = []
        self.verbose = False

    def on_sim_time_sec(self, data):
        if self.start_time is None:
            self.start_time = data
        if self.finish_time is not None and data - self.start_time > self.finish_time:
            self.request_stop()

    def on_robot_xyz(self, data):
        name, arr = data
        self.robot_positions[name] = arr
        if self.verbose:
            self.debug_arr.append((name, arr))

    def on_artf_xyz(self, data):
        for arr in data:
            if arr not in self.artifacts:
                print(self.time, 'received:', arr)
                self.artifacts.append(arr)

    def update(self):
        channel = super().update()

        handler = getattr(self, "on_" + channel, None)
        if handler is not None:
            handler(getattr(self, channel))
        else:
            assert False, channel  # not supported channel

    def draw(self):
        import matplotlib.pyplot as plt
        robot_ids = sorted(list(set([a[0] for a in self.debug_arr])))
        print('Robot IDs', robot_ids)

        for robot_id in robot_ids:
            x = [a[1][1][0] for a in self.debug_arr if a[0] == robot_id]
            y = [a[1][1][1] for a in self.debug_arr if a[0] == robot_id]
            line = plt.plot(x, y, '-o', linewidth=2, label=robot_id)

        artf_types = set([artf[0] for artf in self.artifacts])
        for artf_type in artf_types:
            pos = [pos for artf, pos, src, scored in self.artifacts if artf == artf_type]
            arr_x = [x/1000.0 for x, y, z in pos]
            arr_y = [y/1000.0 for x, y, z in pos]
            plt.plot(arr_x, arr_y, 'x', linewidth=2, label=artf_type)

        #    plt.xlabel('time (s)')
        plt.axes().set_aspect('equal', 'datalim')
        plt.legend()
        plt.show()


def main():
    import argparse
    import os
    from osgar.lib.config import config_load
    from osgar.record import record

    parser = argparse.ArgumentParser(description='SubT Teambase')
    parser.add_argument('config', nargs='+', help='configuration file')
    parser.add_argument('--note', help='add description')
    parser.add_argument('--log', nargs='?', help='record log filename')
    parser.add_argument('--robot-name', required=True, help='T<simulated seconds>F* name')
    args = parser.parse_args()

    cfg = config_load(*args.config)
    cfg['robot']['modules']['app']['init']['robot_name'] = args.robot_name
    prefix = os.path.basename(args.config[0]).split('.')[0] + '-'
    record(cfg, prefix, args.log)


if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4



"""
  Create map (gridmap image)
"""
import os.path

import cv2  # for video output
import numpy as np  # faster depth data processing

from osgar.logger import LogReader, lookup_stream_names, lookup_stream_id
from osgar.lib.serialize import deserialize

SCALE = 10  # 1 pixel is 1dm
BORDER_PX = 10  # extra border


def pts2image(pts, artf, out_filename):
    arr = np.array(pts)
    min_x, min_y = np.min(arr, axis=0)
    max_x, max_y = np.max(arr, axis=0)

    if len(artf) > 0:
        arr_x = [x for __, x, __, __ in artf]
        arr_y = [y for __, __, y, __ in artf]
        min_x = min(min_x, min(arr_x))
        max_x = max(max_x, max(arr_x))
        min_y = min(min_y, min(arr_y))
        max_y = max(max_y, max(arr_y))

    width_px = 2*BORDER_PX + int(SCALE*(max_x - min_x)/1000.0)
    height_px = 2*BORDER_PX + int(SCALE*(max_y - min_y)/1000.0)
    print(width_px, height_px)
    world = np.zeros((height_px, width_px), dtype=np.uint8)
    for x, y in arr:
        px = int(SCALE*(x - min_x)/1000.0) + BORDER_PX
        py = int(SCALE*(y - min_y)/1000.0) + BORDER_PX
        world[height_px - py - 1, px] = 255
    for name, x, y, z in artf:
        px = int(SCALE*(x - min_x)/1000.0) + BORDER_PX
        py = int(SCALE*(y - min_y)/1000.0) + BORDER_PX
        world[height_px - py - 1, px] = 1
        world[height_px - py - 2, px - 1] = 1
        world[height_px - py - 2, px + 1] = 1
        world[height_px - py, px - 1] = 1
        world[height_px - py, px + 1] = 1

    user_color_map = np.zeros((256, 1, 3), dtype=np.uint8)
    user_color_map[0] = (0, 0, 0)
    user_color_map[1] = (255, 255, 255)
    user_color_map[255] = (50, 50, 255)  # BGR -> Red
    cimg = cv2.applyColorMap(world, user_color_map)
    cv2.imwrite(out_filename, cimg)


def main():
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('logfile', help='recorded log file', nargs='+')
    parser.add_argument('--out', help='map output file')
    parser.add_argument('--pose2d', help='stream ID for pose2d messages', default='app.pose2d')
    parser.add_argument('--artf', help='stream ID with artifacts XYZ')
    args = parser.parse_args()

    out_filename = args.out
    if out_filename is None:
        out_filename = os.path.splitext(args.logfile[0])[0] + '.jpg'

    pts = []
    artf = []
    for filename in args.logfile:
        pose_id = lookup_stream_id(filename, args.pose2d)
        streams = [pose_id]
        artf_id = None
        if args.artf is not None:
            artf_id = lookup_stream_id(filename, args.artf)
            streams.append(artf_id)
        for dt, channel, data in LogReader(filename, only_stream_id=streams):
            if channel == pose_id:
                pose = deserialize(data)
                pts.append(pose[:2])
            elif channel == artf_id:
                arr = deserialize(data)
                artf.extend(arr)

    pts2image(pts, artf, out_filename)


if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4 



"""
  SICK Tim571 LIDAR Driver
"""
import numpy as np

from threading import Thread

from osgar.bus import BusShutdownException


STX = b'\x02'
ETX = b'\x03'


class SICKLidar(Thread):
    def __init__(self, config, bus):
        bus.register('raw', 'scan')
        Thread.__init__(self)
        self.setDaemon(True)

        self.bus = bus
        self.buf = b''
        self.sleep = config.get('sleep')
        self.mask = config.get('mask')
        self.blind_zone = config.get('blind_zone')

    @staticmethod
    def parse_raw_data(raw_data):
        """Parse scan data for TiM571, TiM551, TiM881 SICK LIDARs"""
        data = raw_data.split()
        assert len(data) in [26, 854, 846, 583, 1663, 580, 843], len(data)
        assert data[1] == b'LMDscandata', data[:2]
        timestamp = int(data[9], 16)  # TODO verify
        freq = int(data[16], 16)
        if len(data) == 26:  # empty scan
            return [], None
        assert freq == 1500, freq  # TiM 15Hz
        assert data[20] == b'DIST1', data[20]
        resolution = int(data[24], 16)
        assert resolution in [3333, 10000], resolution
        scan_size = int(data[25], 16)
        assert scan_size in [811, 271], scan_size
        scan_start = 26
        scan_end = scan_start + scan_size
        dist = [int(x, 16) for x in data[scan_start:scan_end]]

        remission = None
        # should have have scan remission?
        rssi_index = scan_end + 1
        if rssi_index < len(data) and data[rssi_index] == b'RSSI1':
            assert int(data[rssi_index + 2], 16) == 0, data[rssi_index + 2]
            angular_step = int(data[rssi_index + 4], 16)
            assert angular_step == resolution, (angular_step, resolution)

            amount = int(data[rssi_index + 5], 16)
            assert amount == scan_size, (amount, scan_size)
            rssi_index += 6
            remission = [int(x,16) for x in data[rssi_index:rssi_index + amount]]

        return dist, remission

    def process_packet(self, packet):
        if packet.startswith(STX) and packet.endswith(ETX):
            return self.parse_raw_data(packet)
        return None

    def split_buffer(self, data):
        i = data.find(ETX)
        if i >= 0:
            return data[i + 1:], data[:i + 1]
        return data, b''

    def process_gen(self, data):
        self.buf, packet = self.split_buffer(self.buf + data)
        while len(packet) > 0:
            ret = self.process_packet(packet)
            if ret is not None:
                yield ret
            # now process only existing (remaining) buffer
            self.buf, packet = self.split_buffer(self.buf)  

    def apply_mask(self, scan):
        if self.blind_zone is not None:
            arr = np.array(scan)
            mask = arr < self.blind_zone
            arr[mask] = 0
            scan = arr.tolist()

        if self.mask is None:
            return scan
        assert len(self.mask) == 2, self.mask
        begin, end = self.mask
        assert begin >= 0 and end < 0, (begin, end)  # only index array from begin and end is now supported
        return ([0] * begin) + scan[begin:end] + ([0] * abs(end))

    def run(self):
        try:
            self.bus.publish('raw', STX + b'sRN LMDscandata' + ETX)
            while True:
                packet = self.bus.listen()
                dt, __, data = packet
                for out in self.process_gen(data):
                    assert out is not None
                    scan, remission = out
                    if len(scan) > 0:
                        self.bus.publish('scan', self.apply_mask(scan))
#                    print(dt, [x for x in zip(scan, remission)])
#                    print()
                    if self.sleep is not None:
                        self.bus.sleep(self.sleep)
                    self.bus.publish('raw', STX + b'sRN LMDscandata' + ETX)
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()


# vim: expandtab sw=4 ts=4



"""
  Driver for CO2 sensor via USB, using communication protocol similar to Robik
"""

import struct

from osgar.node import Node
from osgar.bus import BusShutdownException


class MeasureCO2(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('raw', 'co2')
        self.sleep_time = config.get('sleep')
        self._buf = b''

    def query_version(self):
        ret = bytes([0, 0, 3, 0x1, 0x01])
        checksum = sum(ret) & 0xFF
        return ret + bytes([(256-checksum) & 0xFF])

    def create_packet(self):
        # request CO2 readings
        return bytes([0x00,0x00,0x03,0x01,0xC0,0x3C])

    def get_packet(self):
        """extract packet from internal buffer (if available otherwise return None"""
        data = self._buf
        if len(data) < 3:
            return None
        high, mid, low = data[:3]  # 24bit packet length (big endian int)
        assert high == 0, high  # all messages < 65535 bytes
        size = 256 * mid + low + 3  # counting also 3 bytes of packet length header
        if len(data) < size:
            return None
        ret, self._buf = data[:size], data[size:]
        checksum = sum(ret) & 0xFF
        assert checksum == 0, checksum  # checksum error
        return ret

    def parse_CO2_packet(self, data):
        """
        Parse cortexpilot sensors status message
        """
        # expects already validated single sample with 3 bytes length prefix
        #   and checksum at the end
        high, mid, low = data[:3]
        addr, cmd = data[3:5]
        if cmd != 0xC0:
            return None

        assert high == 0, high  # fixed packet size 2*256+89 bytes
        assert mid == 0, mid
        assert low == 7, low
        assert addr == 1, addr
        offset = 5  # payload offset

        return struct.unpack_from('<H', data, offset)[0]

    def run(self):
        try:
            self.publish('raw', self.query_version())
            while True:
                dt, channel, data = self.listen()
                self.time = dt
                if channel == 'raw':
                    self._buf += data
                    packet = self.get_packet()
                    if packet is not None:
                        value_CO2 = self.parse_CO2_packet(packet)
                        if value_CO2 is None:
                            print(packet)
                        else:
                            self.publish('co2', value_CO2)

                        if self.sleep_time is not None:
                            self.sleep(self.sleep_time)
                        self.publish('raw', self.create_packet())

        except BusShutdownException:
            pass

# vim: expandtab sw=4 ts=4



"""
  Driver for robot Eduro
"""


import ctypes
import struct
import math
from threading import Thread

from osgar.bus import BusShutdownException

CAN_ID_SYNC = 0x80
CAN_ID_ENCODERS_LEFT = 0x181
CAN_ID_ENCODERS_RIGHT = 0x182
CAM_SYSTEM_STATUS = 0x8A  # including emergency STOP button
CAN_ID_BUTTONS = 0x28A
CAN_ID_VOLTAGE = 0x18B

UPDATE_TIME_FREQUENCY = 20.0  # Hz
WHEEL_DIAMETER_LEFT = 427.0 / 445.0 * 0.26/4.0 + 0.00015
WHEEL_DIAMETER_RIGHT = 427.0 / 445.0 * 0.26/4.0 - 0.00015
WHEEL_DISTANCE = 0.315


def sint32_diff(a, b):
    return ctypes.c_int32(a - b).value


def CAN_triplet(msg_id, data):
    return [msg_id, bytes(data), 0]  # flags=0, i.e basic addressing


class Eduro(Thread):
    def __init__(self, config, bus):
        bus.register('can', 'buttons', 'voltage', 'emergency_stop', 'encoders', 'pose2d')
        Thread.__init__(self)
        self.setDaemon(True)

        self.bus = bus
        self.time = None

        self.desired_speed = 0.0  # m/s
        self.desired_angular_speed = 0.0
        self.maxAcc = config.get('max_acc', 0.5)
        self._rampLastLeft, self._rampLastRight = 0.0, 0.0

        self.prev_enc_raw = {}
        self.dist_left_diff = 0
        self.dist_right_diff = 0


        self.emergency_stop = None  # uknown state
        self.pose = (0.0, 0.0, 0.0)  # x, y in meters, heading in radians (not corrected to 2PI)
        self.buttons = None  # unknown

    def update_encoders(self, msg_id, data):
        assert len(data) == 4, data
        arr = struct.unpack('<i', data)
        if msg_id in self.prev_enc_raw:
            diff = sint32_diff(arr[0], self.prev_enc_raw[msg_id])

            if msg_id == CAN_ID_ENCODERS_LEFT:
                self.dist_left_diff = diff
            elif msg_id == CAN_ID_ENCODERS_RIGHT:
                self.dist_right_diff = diff
        self.prev_enc_raw[msg_id] = arr[0]

    def update_emergency_stop(self, data):
        assert len(data) == 8, len(data)
        self.emergency_stop = (data[:2] == bytes([0,0x10])) and (data [3:] == bytes([0,0,0,0,0]))

    def update_pose(self):
        x, y, heading = self.pose

        scaleR = WHEEL_DIAMETER_RIGHT * math.pi/0x10000
        scaleL = WHEEL_DIAMETER_LEFT * math.pi/0x10000

        metricL = scaleL * self.dist_left_diff
        metricR = scaleR * self.dist_right_diff

        dist = (metricL + metricR)/2.0
        angle = (metricR - metricL)/WHEEL_DISTANCE

        # advance robot by given distance and angle
        if abs(angle) < 0.0000001:  # EPS
            # Straight movement - a special case
            x += dist * math.cos(heading)
            y += dist * math.sin(heading)
            #Not needed: heading += angle
        else:
            # Arc
            r = dist / angle
            x += -r * math.sin(heading) + r * math.sin(heading + angle)
            y += +r * math.cos(heading) - r * math.cos(heading + angle)
            heading += angle # not normalized
        self.pose = (x, y, heading)

    def send_leds(self, blue, red):
        self.bus.publish('can', CAN_triplet(0x30A, [0xFF, 0xFF, blue, red]))

    def update_buttons(self, data):
        assert len(data) == 2, len(data)
        val = struct.unpack('<H', data)[0] & 0x300
        if self.buttons is None or val != self.buttons:
            self.buttons = val
            msg = { 'blue_selected': ((val & 0x0200) == 0),
                    'cable_in': ((val & 0x0100) != 0) }
            if msg['blue_selected']:
                self.send_leds(blue=1, red=0)
            else:
                self.send_leds(blue=0, red=1)
            self.bus.publish('buttons', msg)

    def update_voltage(self, data):
        battery = data[1] * 256 + data[0]  # 1/100th volts
        self.bus.publish('voltage', [battery])

    def send_speed(self):
        self.SpeedL = self.desired_speed - self.desired_angular_speed * WHEEL_DISTANCE / 2.0
        self.SpeedR = self.desired_speed + self.desired_angular_speed * WHEEL_DISTANCE / 2.0 

        if False: #any(motorId in self.modulesForRestart for motorId in [0x01, 0x02]):
            # There is a motor in reset => we must stop (even aggressively)
            left, right = 0, 0
        else:
            scaleR = 1000.0 / ( math.pi * WHEEL_DIAMETER_RIGHT)
            scaleL = 1000.0 / ( math.pi * WHEEL_DIAMETER_LEFT )
            tmpSpeedL = self.SpeedL
            tmpSpeedR = self.SpeedR

        if self.maxAcc:
            # use ramps
            maxSpeedStep = self.maxAcc / UPDATE_TIME_FREQUENCY;
            if math.fabs( tmpSpeedL - self._rampLastLeft ) > maxSpeedStep or \
                math.fabs( tmpSpeedR - self._rampLastRight ) > maxSpeedStep:
                frac = maxSpeedStep / max( math.fabs( tmpSpeedL - self._rampLastLeft ), math.fabs( tmpSpeedR - self._rampLastRight ) )
                tmpSpeedL = self._rampLastLeft + frac * ( tmpSpeedL - self._rampLastLeft )
                tmpSpeedR = self._rampLastRight + frac * ( tmpSpeedR - self._rampLastRight )
                self._rampLastLeft = tmpSpeedL
                self._rampLastRight = tmpSpeedR

        left = int(scaleL * tmpSpeedL)
        right = int(scaleR * tmpSpeedR)

        maxLim = 4000
        if left > maxLim:
            right = right*maxLim//left
            left = maxLim
        if left < -maxLim:
            right = -right*maxLim//left
            left = -maxLim
        if right > maxLim:
            left = left*maxLim//right
            right = maxLim
        if right < -maxLim:
            left = -left*maxLim//right
            right = -maxLim

        self.bus.publish('can', CAN_triplet(0x201, [
            left&0xff, (left>>8)&0xff,
            right&0xff, (right>>8)&0xff]))

    def send_speed_ref(self):
        if self.desired_speed > 0:
            left, right = 1000, 1000
        else:
            left, right = 0, 0
        self.bus.publish('can', CAN_triplet(0x201, [
            left&0xff, (left>>8)&0xff,
            right&0xff, (right>>8)&0xff]))

    def check_restarted_modules(self, module_id, status):
        if module_id in [1, 2]:  # motors
            if status != 5:
                print(self.time, module_id, status)
                self.prev_enc_raw = {}
                self.dist_left_diff = 0
                self.dist_right_diff = 0

    def process_packet(self, packet, verbose=False):
        msg_id, payload, flags = packet
        if msg_id in [CAN_ID_ENCODERS_LEFT, CAN_ID_ENCODERS_RIGHT]:
            self.update_encoders(msg_id, payload)
        elif msg_id == CAM_SYSTEM_STATUS:
            self.update_emergency_stop(payload)
            self.bus.publish('emergency_stop', self.emergency_stop)
            if self.emergency_stop:
                print("Eduro - emergency stop")
                self.bus.publish('can', CAN_triplet(0, [0x80, 0]))  # sendPreOperationMode
        elif msg_id == CAN_ID_BUTTONS:
            self.update_buttons(payload)
        elif msg_id == CAN_ID_VOLTAGE:
            self.update_voltage(payload)
        elif msg_id == CAN_ID_SYNC:
            # make sure diff is reported _before_ update_pose() which reset counters!
            self.bus.publish('encoders', [self.dist_left_diff, self.dist_right_diff])
            self.update_pose()
            x, y, heading = self.pose
            self.bus.publish('pose2d', [round(x*1000), round(y*1000), round(math.degrees(heading)*100)])
            self.send_speed()
        elif msg_id & 0xFF0 == 0x700:  # heart beat message
            assert len(payload) == 1, len(payload)
            self.check_restarted_modules(msg_id & 0xF, payload[0])

    def process_gen(self, data, verbose=False):
        self.process_packet(data)
        yield None

    def slot_can(self, timestamp, data):
        if len(data) > 0:
            for status in self.process_gen(data):
                if status is not None:
                    self.bus.publish('can', status)  # TODO

    def slot_desired_speed(self, timestamp, data):
        self.desired_speed, self.desired_angular_speed = data[0]/1000.0, math.radians(data[1]/100.0)

    def run(self):
        try:
            while True:
                self.time, channel, data = self.bus.listen()
                if channel == 'can':
                    self.slot_can(self.time, data)
                elif channel == 'desired_speed':
                    self.slot_desired_speed(self.time, data)
                else:
                    assert False, channel  # unsupported channel
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()

# vim: expandtab sw=4 ts=4



"""
  Explore - follow a wall/obstacles
    (use 2D SICK LIDAR only)
"""
import math
from datetime import timedelta

import numpy as np

from osgar.node import Node


WALL_DISTANCE = 1.5  #0.7  # m
DESIRED_SPEED = 0.5  # m/s

BLIND_ZONE_MM = 100  # self-reflections of the laser itself


def min_dist(data):
    data = np.array(data)
    mask = (data > BLIND_ZONE_MM)  # reject ultra close reflections and 0 = no reflection
    if np.any(mask):
        return np.min(data[mask]) * 0.001
    return None 


def tangent_circle(dist, radius):
    """
      return tangent angle to a circle placed at (dist, 0.0) with radius=radius

      For non-existing tangent use 100 degrees.
    """
    if dist >= radius:
        return math.asin(radius/float(dist))
    return math.radians(100)


def follow_wall_angle(laser_data, radius, right_wall=False):
    """
        Find the angle to the closest point in laser scan (either on the left or right side).
        Then calculate an angle to a free space as tangent to circle of given radius.
        This angle is returned and can be used for steering command.
    """
    data = np.array(laser_data)
    size = len(laser_data)
    deg_resolution = 270 / (size - 1)  # SICK uses extra the first and the last, i.e. 271 rays for 1 degree resolution
    mask = (data <= 300)  # ignore internal reflections
    data[mask] = 20000

    mask = (data >= 4000)  # ignore obstacles beyond 4m
    data[mask] = 20000

    # To make the code simpler, let's pretend we follow the right wall and flip
    # the result in the end if we are actually following the left wall.
    if not right_wall:
        data = data[::-1]

    distances = data / 1000.0

    r = radius * 1  # TODO: Or some other first guess?
    INCREMENT = 0.3  # In meters. TODO: Some other value, maybe?
    found_wall = False
    for attempt in range(50):
        # Find where the wall we follow starts.
        wall_start_idx = -1
        # We only accept walls to the right of the robot.
        for (i, distance) in enumerate(distances[:size//2]):
            if distance <= r:
                wall_start_idx = i
                break
        if wall_start_idx < 0:
            r += INCREMENT
        else:
            found_wall = True
            break

    if not found_wall:
        # No wall found. Let's slowly circle.
        # TODO: ideally, this would be stateful and we would spiral.
        return math.radians(-20 if right_wall else 20)

    last_wall_idx = wall_start_idx
    while True:
        last_wall_distance = distances[last_wall_idx]
        found_countinuation = False
        for i in range(last_wall_idx + 1, size):
            dist = distances[i]
            rel_idx = i - last_wall_idx
            sin_angle = math.sin(rel_idx * math.radians(deg_resolution))
            cos_angle = math.cos(rel_idx * math.radians(deg_resolution))
            # How far is the currently observed point from the previous wall point?
            gap = math.hypot(cos_angle * dist - last_wall_distance,
                             sin_angle * dist - 0)
            if gap <= radius:
                last_wall_idx = i
                found_countinuation = True
                break
        if not found_countinuation:
            break

    # If we do not see the end of the wall because of occlusion, our desired
    # direction is just towards the last wall point we see. Only otherwise we
    # can keep a safe distance from the wall we follow.
    last_wall_distance = distances[last_wall_idx]
    if last_wall_idx + 1 < size and distances[last_wall_idx + 1] < last_wall_distance:
        tangent_angle = 0.
    else:
        tangent_angle = tangent_circle(last_wall_distance, radius)


    laser_angle = math.radians(-135 + last_wall_idx * deg_resolution)
    total_angle = laser_angle + tangent_angle
    if not right_wall:
        total_angle = -total_angle

    return total_angle


class FollowWall(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('desired_speed')
        self.right_wall = config.get('right_wall', False)
        self.max_speed = DESIRED_SPEED

    def send_speed_cmd(self, speed, angular_speed):
        return self.publish('desired_speed', [round(speed*1000), round(math.degrees(angular_speed)*100)])

    def update(self):  # hack, this method should be called run instead!       
        channel = super().update()  # define self.time

        desired_speed = 0.0
        start_time = self.time
        while self.time - start_time < timedelta(minutes=1):
            channel = super().update()
            if channel == 'scan':
                size = len(self.scan)
                dist = min_dist(self.scan[size//3:2*size//3])
                tangent_angle = follow_wall_angle(self.scan, right_wall=self.right_wall,
                                                  radius=WALL_DISTANCE)
                if dist < 2.0:
                    desired_speed = (self.max_speed / 2) * (dist - 0.4) / 1.6
                else:
                    desired_speed = self.max_speed
                desired_angular_speed = 0.7 * tangent_angle  # TODO better P-regulator based on angle and free space
                self.send_speed_cmd(desired_speed, desired_angular_speed)

            elif channel == 'emergency_stop':
                self.send_speed_cmd(0.0, 0.0)
                self.request_stop()  # it should be "delayed"


if __name__ == "__main__":
    from osgar.launcher import launch

    launch(app=FollowWall, description='Follow Wall', prefix='wall-')

# vim: expandtab sw=4 ts=4 




"""
  Virtual SubT Challenge - decoder of robot name

The name is based on combination of capital letters and numbers.
New coding:
  <name letter (any)> ([<duration>] <action type>)+
  where supported actions are (first letter) Left, Right, Center, Wait, Home

The last action is by default Home. There is no default direction.
The time for Home is optional (default is 2 times sum of previous exploration).
"""

def split_multi(s, delimiters):
    # split string s via multiple character delimiters
    ret = []
    word = ''
    for c in s:
        word += c
        if c in delimiters:
            ret.append(word)
            word = ''
    if word != '':
        ret.append(word)
    return ret


def parse_robot_name(robot_name):
    options = {
        'L': 'left',
        'R': 'right',
        'C': 'center',
        'W': 'wait',
        'H': 'home'
    }
    parts = split_multi(robot_name[1:], options.keys())
    ret = []
    sum_t = 0
    entered = False
    for part in parts:
        action = options[part[-1]]
        if len(part) > 1:
            t = int(part[:-1])
        else:
            assert action == 'home', action  # missing time
            t = 2 * sum_t
        if action in ['left', 'right', 'center']:
            # exploration commands
            if not entered:
                ret.append(('enter-' + action, None))
                entered = True
            sum_t += t
        ret.append((action, t))
        if action == 'home':
            entered = False
            sum_t = 0
    if len(ret) > 0 and ret[-1][0] != 'home':
        # add default home return if it was not specifically requested
        ret.append(('home', sum_t * 2))
    return ret

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock
import math

import numpy as np

from subt.points2scan import PointsToScan


class PointsToScanTest(unittest.TestCase):

    def test_numpy(self):
        bus = MagicMock()
        conv = PointsToScan(bus=bus, config={})

        orig_data = np.zeros((16, 10000, 3), dtype=np.float32)
        data = orig_data[:, ::10, :]  # downsample to everh 10th
        conv.on_points(data)
        bus.publish.assert_called()

# vim: expandtab sw=4 ts=4




"""
  Loop detection.
"""

import math

from osgar.lib.quaternion import angle_between

class LoopDetector:
    def __init__(self, orientation_similarity_threshold=math.radians(20),
                       pose_distance_threshold=0.5,
                       granularity=0.3,
                       max_loop_length=100.0,
                       min_loop_length=10.0):
        self.orientation_similarity_threshold = orientation_similarity_threshold
        self.pose_distance_threshold = pose_distance_threshold
        self.granularity = granularity
        self.max_loop_length = max_loop_length
        self.min_loop_length = min_loop_length
        self.trajectory = []
        self.detected_loop = []


    def add(self, xyz, orientation):
        if not self.trajectory:
            self.trajectory.append([xyz, orientation, 0])
            return

        prev_xyz = self.trajectory[-1][0]
        d = math.sqrt(sum((a - b)**2 for (a, b) in zip(xyz, prev_xyz)))
        if d < self.granularity:
            return
        self.trajectory[-1][2] = d
        self.trajectory.append([xyz, orientation, 0])

        self.detected_loop = self.__detect_loop()


    def add_all(self, poses):
        for pose in poses:
            self.add(*pose)


    def __detect_loop(self):
        loop_candidate = []
        loop_candidate_length = 0.0
        is_loop = False

        curr_xyz, curr_orientation, _ = self.trajectory[-1]

        for past_xyz, past_orientation, d in reversed(self.trajectory):
            loop_candidate.append((past_xyz, past_orientation))

            loop_candidate_length += d
            if loop_candidate_length < self.min_loop_length:
                continue

            if loop_candidate_length > self.max_loop_length:
                break

            dist = math.sqrt(sum((a-b)**2 for (a, b) in zip(curr_xyz, past_xyz)))
            if dist > self.pose_distance_threshold:
                continue

            angle = angle_between(curr_orientation, past_orientation)
            if angle > self.orientation_similarity_threshold:
                continue

            is_loop = True
            break


        return loop_candidate[::-1] if is_loop else []

    def loop(self):
        return self.detected_loop if self.detected_loop else None

if __name__ == '__main__':
    import argparse
    import cv2
    import os
    import sys

    from osgar.lib.serialize import deserialize
    from osgar.logger import LogReader, lookup_stream_id

    parser = argparse.ArgumentParser()
    parser.add_argument('--logfile', help='Path to Osgar log.', required=True)
    parser.add_argument('--pose3d-stream',
                        help='Stream id or name with pose and orientation data.',
                        required=True)
    parser.add_argument('--debug', help='Provide debug info.', action='store_true')
    args = parser.parse_args()

    pose3d_stream_id = lookup_stream_id(args.logfile, args.pose3d_stream)
    loop_detector = LoopDetector()
    with LogReader(args.logfile,
                   only_stream_id=pose3d_stream_id) as logreader:
        last_detected_start = None
        for time, stream, data in logreader:
            pose, orientation = deserialize(data)
            loop_detector.add(pose, orientation)
            loop = loop_detector.loop()
            if loop:
                if args.debug:
                    print('xxxx', loop)
                else:
                    if last_detected_start is None:
                        last_detected_start = time
            else:
                if last_detected_start is not None:
                    print(f'Loop detected at {last_detected_start} (duration {time - last_detected_start})')
                    last_detected_start = None




"""
  Wrapper for PeakCAN USB device
"""
#
# pip install python-can
# https://python-can.readthedocs.io/en/2.1.0/interfaces/pcan.html
#
# https://www.peak-system.com/
# https://www.peak-system.com/fileadmin/media/files/pcan-basic.zip
# http://www.peak-system.com/quick/BasicLinux
#
# https://en.wikipedia.org/wiki/CAN_bus
#
# The new CAN messages will contain arbitration_id, data and flags
#  http://docs.ros.org/melodic/api/can_msgs/html/msg/Frame.html
#

import can
from threading import Thread

from osgar.bus import BusShutdownException

IS_EXTENDED_ID_MASK = 0x1

class PeakCAN:
    def __init__(self, config, bus):
        self.bus = bus
        bus.register('can')
        self.canbus = can.interface.Bus(bustype='pcan', channel='PCAN_USBBUS1', bitrate=500000)
        self.input_thread = Thread(target=self.run_input, daemon=True)
        self.output_thread = Thread(target=self.run_output, daemon=True)

    def start(self):
        self.input_thread.start()
        self.output_thread.start()

    def join(self, timeout=None):
        self.input_thread.join(timeout=timeout)
        self.output_thread.join(timeout=timeout)

    def run_input(self):
        while self.bus.is_alive():
            msg = self.canbus.recv()  # TODO timeout
            flags = IS_EXTENDED_ID_MASK if msg.is_extended_id else 0
            self.bus.publish('can', [msg.arbitration_id, msg.data, flags])

    def slot_raw(self, timestamp, packet):
        arbitration_id, data, flags = packet
        msg = can.Message(arbitration_id=arbitration_id, data=data, is_extended_id=(flags & IS_EXTENDED_ID_MASK))
        self.canbus.send(msg)  # TODO timeout, locks?!

    def run_output(self):
        try:
            while True:
                dt, __, data = self.bus.listen()
                self.slot_raw(dt, data)
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()


# vim: expandtab sw=4 ts=4



import collections
import datetime
import inspect
import logging
import math

from threading import Thread, RLock

import numpy as np
from shapely.geometry import LineString, MultiLineString

from osgar.lib import quaternion


g_logger = logging.getLogger(__name__)


def log(*args):
    up = inspect.currentframe().f_back
    caller = up.f_code.co_name
    caller_class = up.f_locals['self'].__class__
    msg = f"{caller_class.__name__}.{caller}, " + ", ".join(str(a) for a in args)
    g_logger.info(msg)


class SimLogger:

    def __init__(self):
        self.lock = RLock()
        self.start_time = datetime.datetime.now(datetime.timezone.utc)
        self.names = ["sys"]

    def register(self, name):
        self.names.append(name)
        return len(self.names)-1

    def write(self, stream_id, data, dt=None):
        dt = datetime.datetime.now(datetime.timezone.utc) - self.start_time
        return dt


class Simulation:

    def __init__(self, bus, end_condition, world=None):
        bus.register('scan', 'rot', 'orientation', 'sim_time_sec', 'origin', 'pose2d', 'acc', 'artf', 'pose3d')
        self.bus = bus
        self.end_condition = end_condition
        self.world = world if world is not None else MultiLineString()
        self._handlers = collections.defaultdict(lambda: self.on_default)
        self.set_handler('request_origin', self.on_request_origin)
        self.set_handler('desired_speed', self.on_desired_speed)
        self.set_handler('pose2d', self.on_pose)
        self.set_handler('pose3d', self.on_pose)
        self.origin = [7.0, 3.0, 0.0]
        self.xyz = [0.0, 0.0, 0.0]
        self.orientation = quaternion.identity()
        self.speed_forward = 0.0
        self.speed_angular = 0.0
        self.time = datetime.timedelta()

    def start(self):
        self.thread = Thread(target=self.main)
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout)

    def main(self):
        log("started")
        self.on_request_origin(datetime.timedelta(), "request_origin", True)
        # there is no SubT control until pose is received -> trigger it
        self.on_desired_speed(datetime.timedelta(), "desired_speed", [0, 0])

        for _ in range(500):
            packet = self.bus.listen()
            self.handle_incomming(*packet)
            if self.end_condition(self):
                return

    def set_handler(self, channel, handler):
        self._handlers[channel] = handler

    def handle_incomming(self, dt, channel, data):
        self._handlers[channel](dt, channel, data)

    def on_default(self, dt, channel, data):
        log(dt, channel, data)

    def on_request_origin(self, dt, channel, data):
        log(dt, channel, data)
        # ros_proxy_node.cc keeps sending origin data until robot arrives within 0.3m from (0,0,0)
        # we send it only once
        corrected = [rr - oo for rr, oo in zip(self.xyz, self.origin)]
        self.bus.publish('origin', [b'X100L', *corrected, *self.orientation])
        self.bus.publish('scan', self.scan())
        heading_centidegrees = round(math.degrees(quaternion.heading(self.orientation))*100)
        self.bus.publish('rot', [heading_centidegrees, 0, 0])
        self.bus.publish('orientation', self.orientation)

    def on_pose(self, dt, channel, data):
        if channel == 'pose2d':
            out = str(data)
        elif channel == 'pose3d':
            out0 = "[" + ', '.join([f"{p:.2f}" for p in data[0]]) + "]"
            out1 = " [" + ', '.join([f"{p:.2f}" for p in data[1]]) + "]"
            out = out0 + out1
        log(dt, channel, out)

    def on_desired_speed(self, dt, channel, data):
        forward = data[0]/1000.               # from millimeters per second to meters per second
        angular = math.radians(data[1]/100.)  # from centidegrees per second to radians per second
        dtime = 0.1                           # simulation step time in seconds
        max_acc_forward = 1                   # meters per sec**2
        max_acc_angular = 1                   # radians per sec**2
        acc_forward = np.clip(-max_acc_forward, (forward - self.speed_forward)/dtime, max_acc_forward)
        acc_angular = np.clip(-max_acc_angular, (angular - self.speed_angular)/dtime, max_acc_angular)
        self.speed_forward += acc_forward * dtime
        self.speed_angular += acc_angular * dtime
        log("desired", f"forward {forward:.2f} m/s", f"angular {math.degrees(angular):.2f} deg/s")
        log("current", f"forward {self.speed_forward:.2f} m/s", f"angular {math.degrees(self.speed_angular):.2f} deg/s")
        dist = dtime * self.speed_forward
        dist3d = quaternion.rotate_vector([dist, 0, 0], self.orientation)
        self.xyz = [a + b for a, b in zip(self.xyz, dist3d)]
        turn = quaternion.from_axis_angle((0,0,1), dtime * self.speed_angular)
        self.orientation = quaternion.multiply(self.orientation, turn)
        self.time += datetime.timedelta(seconds=dtime)
        heading_centidegrees = round(math.degrees(quaternion.heading(self.orientation))*100)
        x_mm, y_mm, z_mm = [round(c*1000) for c in self.xyz]
        self.bus.publish('orientation', self.orientation)
        self.bus.publish('rot', [heading_centidegrees, 0, 0])
        self.bus.publish('pose2d', [x_mm, y_mm, heading_centidegrees])
        self.bus.publish('pose3d', [[a - b for a, b, in zip(self.xyz, self.origin)], self.orientation])
        self.bus.publish('sim_time_sec', self.time.total_seconds())
        self.bus.publish('scan', self.scan())

        #log("++++", x, y, math.degrees(heading))

    def scan(self):
        from shapely.affinity import translate, rotate
        "270 degrees -> np.linspace(-135, 135, 271)"
        ret = np.zeros((271,))

        # create lidar lines
        angles = np.linspace(math.radians(-135), math.radians(135), 271)
        acos = np.cos(angles)
        asin = np.sin(angles)
        max_dist_m = 10.0
        x = acos * max_dist_m
        y = asin * max_dist_m

        # transform world to robot coordinates
        world = self.world
        world = translate(world, *[-c for c in self.xyz])
        world = rotate(world, -quaternion.heading(self.orientation), origin=(0,0), use_radians=True)

        # intersect each lidar line with the world
        for i, point in enumerate(zip(x, y)):
            cross = world.intersection(LineString([(0,0), point]))
            if cross.is_empty:
                continue
            if cross.geom_type == 'Point':
                ret[i] = math.hypot(*cross.coords)
            elif cross.geom_type == 'MultiPoint':
                # find closest intersection
                ret[i] = max_dist_m
                for item in cross:
                    dist = math.hypot(*cross.coords)
                    if ret[i] > dist:
                        ret[i] = dist
            else:
                assert False
        return ret

if __name__ == "__main__":
    from unittest.mock import MagicMock
    from shapely.geometry import MultiLineString
    bus = MagicMock()
    world = MultiLineString([[(1,-1), (1,1)], [(2,-1), (2,1)]])
    sim = Simulation(bus, lambda sim: False, world=world)
    sim.xyz = [-1.0, 0.0, 0.0]
    sim.orientation = quaternion.from_axis_angle((0, 0, 1), math.radians(90))
    scan = sim.scan()
    print(scan)



import os
import unittest
from osgar.lib.config import config_load, merge_dict, MergeConflictError, get_class_by_name
from osgar.drivers.logsocket import LogTCPStaticIP as LogTCP

def test_data(filename, test_dir='test_data'):
    return os.path.join(os.path.dirname(__file__), test_dir, filename)


class MyTestRobot:
    def get_name(self):
        return 'Karel'


class ConfigTest(unittest.TestCase):

    def test_load_config_files(self):
        conf_dir = '../../config'
        filename = test_data('ro2018-spider-gps-imu.json', conf_dir)
        conf = config_load(filename)
        self.assertEqual(conf['version'], 2)

    def test_multiple_config_files(self):
        conf_dir = '../../config'
        filename1 = test_data('ro2018-spider-gps-imu.json', conf_dir)
        filename2 = test_data('ro2018-czu-waypoints.json', conf_dir)
        conf = config_load(filename1, filename2)
        self.assertIn('maxspeed',conf['robot']['modules']['app']['init'])
        self.assertIn('out',conf['robot']['modules']['app'])

    def test_merge_dict(self):
        self.assertEqual(merge_dict({'A':1}, {'B':2}), {'A':1, 'B':2})

        dict1 = {'root':{'A':1}}
        dict2 = {'root':{'B':2}}
        merge = {'root':{'A':1, 'B':2}}
        self.assertEqual(merge_dict(dict1, dict2), merge)

        self.assertEqual(merge_dict({'A':1}, {'A':1}), {'A':1})
        self.assertEqual(merge_dict({'A':1}, {'A':2}), {'A':2})
        self.assertEqual(merge_dict({'A':'a'}, {'A':'b'}), {'A':'b'})
        self.assertEqual(merge_dict({'A':[]}, {'A':[1]}), {'A':[1]})
        self.assertEqual(merge_dict({'A':()}, {'A':(1,)}), {'A':(1,)})

        with self.assertRaises(MergeConflictError) as e:
            merge_dict({'A':{'B':1}}, {'A':2})

    def test_get_class_by_name(self):
        robot = get_class_by_name('osgar.lib.test_config:MyTestRobot')()
        self.assertEqual(robot.get_name(), 'Karel')

        node = get_class_by_name('tcp')
        self.assertEqual(node, LogTCP)

        node = get_class_by_name('udp')
        self.assertNotEqual(node, LogTCP)

    def test_application(self):
        conf_dir = '../../config'
        filename = test_data('ro2018-spider-gps-imu.json', conf_dir)
        conf = config_load(filename, application=MyTestRobot)
        self.assertTrue(conf['robot']['modules']['app']['driver'].endswith('lib.test_config:MyTestRobot'))
        conf = config_load(filename, application='osgar.lib.test_config:MyTestRobot')
        self.assertTrue(conf['robot']['modules']['app']['driver'].endswith('lib.test_config:MyTestRobot'))

# vim: expandtab sw=4 ts=4




import collections
import datetime
import logging
import signal
import subprocess
import sys
import threading
import time


import zmq

import osgar.lib.serialize
import osgar.lib.config
import osgar.logger

ENDPOINT = "tcp://127.0.0.1:8882"
ASSERT_QUEUE_DELAY = datetime.timedelta(seconds=.1)
STOPPING_TIMEOUT = datetime.timedelta(seconds=2)

g_logger = logging.getLogger(sys.modules[__name__].__spec__.name)


def child(name, module_config, log_level):
    # todo how to inherit logging setup from parent process?
    # a proper way is probably using a config file
    import logging, sys
    #print(logging.getLevelName(log_level))
    logging.basicConfig(
        stream=sys.stdout,
        level=log_level,
        format='%(asctime)s %(name)-16s %(levelname)-8s %(message)s',
    )
    # todo end
    signal.signal(signal.SIGINT,signal.SIG_IGN)
    klass = osgar.lib.config.get_class_by_name(module_config['driver'])
    bus = _Bus(name)
    instance = klass(config=module_config.get('init', {}), bus=bus)
    instance.start()
    g_logger.info(f"{name} running")
    instance.join()
    g_logger.info(f"{name} finished")
    bus.request_stop()
    g_logger.debug(f"{name} stop confirmed")


def record(config, log_prefix=None, log_filename=None, duration_sec=None):
    g_logger.info("recording...")
    with osgar.logger.LogWriter(prefix=log_prefix, filename=log_filename, note=str(sys.argv)) as log:
        log.write(0, bytes(str(config), 'ascii'))
        g_logger.info(log.filename)
        with _Router(log) as router:
            modules = {}
            s = sys.modules[__name__].__spec__
            for module_name, module_config in config['robot']['modules'].items():
                program = f"import {s.name}; {s.name}.child('{module_name}', {module_config}, {logging.root.level})"
                modules[module_name] = subprocess.Popen([sys.executable, "-c", program])

            try:
                router.register_nodes(modules.keys(), timeout=datetime.timedelta(seconds=20))
                links =  config['robot']['links']
                for link_from, link_to in links:
                    router.connect(link_from, link_to)
                duration = datetime.timedelta(seconds=duration_sec) if duration_sec else datetime.timedelta.max
                router.run(duration=duration)
            except Exception as e:
                (etype, evalue, tb) = sys.exc_info()
                import traceback
                traceback.print_tb(tb)
                g_logger.error(str(e))
                router.request_stop(b"exception")

            for module in modules.values():
                try:
                    # always stop within 1s
                    timeout = 1 - (router.now() - router.stopping).total_seconds()
                    if timeout < 0:
                        timeout = 0
                    module.wait(timeout)
                except subprocess.TimeoutExpired:
                    module.kill()
                    module.wait()


class _Router:
    def __init__(self, logger):
        self.logger = logger
        self.start_time = self.logger.start_time
        self._context = zmq.Context()
        self.socket = self._context.socket(zmq.ROUTER)
        self.nodes = dict()
        self.delays = dict()
        self.last_listen = dict()
        self.stopped = set()
        self.subscriptions = dict()
        self.stream_id = dict()
        self.listening = set()
        self.stopping = datetime.timedelta()
        self.no_output = set()
        self.compressed_output = set()
        self._saved_sigint = None

    def __enter__(self):
        self.socket.bind(ENDPOINT)
        self._saved_sigint = signal.signal(signal.SIGINT, self.sigint)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.signal(signal.SIGINT, self._saved_sigint)
        self.socket.close()
        self._context.term()

    def sigint(self, signum, frame):
        # TODO se pipe-to-self trick to tell receive loop it is time to quit
        g_logger.error("ctrl-c")
        self.request_stop(b'ctrl-c')

    def register_nodes(self, nodes, timeout=datetime.timedelta.max):
        # wait for all nodes to "register" their outputs
        nodes = set(bytes(node_name, "ascii") for node_name in nodes)
        for packet in self.receive(timeout=timeout):
            sender, action, *args = packet
            assert action == b"register", (sender, action, args)
            assert sender not in self.nodes             # we have not registered the node yet
            assert sender in nodes, (sender, nodes)     # it is one of the nodes we expect
            self.nodes[sender] = collections.deque()    # receiving queue
            self.delays[sender] = datetime.timedelta()
            for name_and_type in args:
                o, *suffix = name_and_type.split(b':')
                suffix = suffix[0] if suffix else b''
                link_from = sender + b"." + o
                idx = self.logger.register(str(link_from, 'ascii'))
                self.stream_id[link_from] = idx
                if suffix == b'null':
                    self.no_output.add(idx)
                elif suffix == b'gz':
                    self.compressed_output.add(idx)
            if self.nodes.keys() == nodes:
                return
        raise RuntimeError("unexpected stop")

    def connect(self, link_from, link_to):
        link_from = bytes(link_from, "ascii")
        link_to = bytes(link_to, "ascii")
        assert link_from in self.stream_id, (link_from, self.stream_id.keys())
        receiver, channel = link_to.split(b".")
        self.subscriptions.setdefault(link_from, []).append((receiver, channel))
        g_logger.info(f"connect {link_from} -> {link_to}")

    def run(self, *, duration=datetime.timedelta.max):
        # answer all connected nodes that it is ok to run now (connections are set up)
        for node_name in self.nodes:
            self.socket.send_multipart([node_name, b"", b"start"])

        for packet in self.receive(duration=duration):
            sender, action, *args = packet
            getattr(self, str(action, 'ascii'))(sender, *args)

        for sender, delay in self.delays.items():
            if delay > ASSERT_QUEUE_DELAY:
                g_logger.error(f"{str(sender, 'ascii')} max delay: {delay}")
            else:
                g_logger.info(f"{str(sender, 'ascii')} max delay: {delay}")

    def receive(self, *, hz=10, timeout=datetime.timedelta.max, duration=datetime.timedelta.max):
        socket = self.socket
        poller = zmq.Poller()
        poller.register(socket, zmq.POLLIN)
        start_time = self.now()
        while True:
            now = self.now()
            running_time = now - start_time
            if running_time > timeout:
                raise RuntimeError("timeout")
            if not self.stopping and running_time > duration:
                self.request_stop(b"duration")
            if self.stopping and now - self.stopping > STOPPING_TIMEOUT:
                g_logger.error(f'failed to stop within timeout of {STOPPING_TIMEOUT}, exiting anyway')
                for name, q in self.nodes.items():
                    if len(q) > 0:
                        g_logger.error(f"{name} queue: {len(q)}")
                g_logger.error(f"still running: {self.nodes.keys() - self.stopped}")
                return
            if self.stopping and len(self.nodes.keys() - self.stopped) == 0:
                g_logger.info('all done, stopping')
                return
            obj = dict(poller.poll(1000//hz))
            if socket in obj and obj[socket] == zmq.POLLIN:
                packet = socket.recv_multipart()
                yield (packet[0], *packet[2:]) # drop frame separator from REQ socket

    def listen(self, sender):
        assert sender.endswith(b'-listen')
        node_name = sender[:-len(b'-listen')]
        queue = self.nodes[node_name]
        if len(queue) == 0:
            self.listening.add(node_name)
        else:
            packet = queue.popleft()[1]
            assert packet[0] == sender, (packet[0], sender)
            self.socket.send_multipart(packet)
            self.last_listen[node_name] = self.now()

    def publish(self, sender, channel, data):
        dt = self.now()
        dt_uint32 = osgar.logger.format_timedelta(dt)
        self.socket.send_multipart([sender, b"", b"publish", dt_uint32])
        link_from = sender + b"." + channel
        for node_name, input_channel in self.subscriptions.get(link_from, []):
            self.send(node_name, b'listen', dt, input_channel, data)
        stream_id = self.stream_id[link_from]
        if stream_id in self.no_output:
            data = osgar.lib.serialize.serialize(None)
        elif stream_id in self.compressed_output:
            data = osgar.lib.serialize.compress(data)
        self.logger.write(stream_id, data, dt)

        delay = dt - self.last_listen.get(sender, dt)
        if self.delays.get(sender, datetime.timedelta()) < delay:
            self.delays[sender] = delay
        if delay > ASSERT_QUEUE_DELAY:
            self.report_error(name=sender, delay=delay.total_seconds(), channel=channel)

    def request_stop(self, sender):
        # confirm
        dt_uint32 = osgar.logger.format_timedelta(self.now())
        packet = [sender, b"", b"quit", dt_uint32]
        self.socket.send_multipart(packet)
        assert sender not in self.stopped
        self.stopped.add(sender)
        if self.stopping:
            return
        g_logger.info(f"{str(sender, 'ascii')} requested stop")
        self.stopping = self.now()
        for node_name in self.nodes:
            if node_name != sender:
                self.send(node_name, b"quit", self.stopping)

    def is_alive(self, sender):
        dt_uint32 = osgar.logger.format_timedelta(self.now())
        self.socket.send_multipart([sender, b"", b"quit" if self.stopping else b"is_alive", dt_uint32])

    def send(self, node_name, action, dt, *args):
        dt_uint32 = osgar.logger.format_timedelta(dt)
        packet = [node_name+b'-listen', b"", action, dt_uint32, *args]
        if node_name in self.listening:
            self.listening.remove(node_name)
            self.socket.send_multipart(packet)
            self.last_listen[node_name] = dt
        else:
            self.nodes[node_name].append((dt, packet))

    def now(self):
        return datetime.datetime.now(datetime.timezone.utc) - self.start_time

    def report_error(self, **err):
        self.logger.write(0, bytes(str(err), encoding='ascii'))
        g_logger.error(str(err))


class _Bus:
    def __init__(self, name):
        self.lock_publish = threading.Lock()
        self.lock_listen = threading.Lock()
        self.name = name
        context = zmq.Context()
        self.sock_publish = context.socket(zmq.REQ)
        self.sock_publish.setsockopt_string(zmq.IDENTITY, name)
        self.sock_publish.connect(ENDPOINT)
        self.sock_listen = context.socket(zmq.REQ)
        self.sock_listen.setsockopt_string(zmq.IDENTITY, name+'-listen')
        self.sock_listen.connect(ENDPOINT)
        self.parse_listen_dt = osgar.logger.timedelta_parser()
        self.parse_publish_dt = osgar.logger.timedelta_parser()

    def register(self, *outputs):
        bytes_outputs = list(bytes(o, 'ascii') for o in outputs)
        with self.lock_publish:
            self.sock_publish.send_multipart([b"register", *bytes_outputs])
            resp = self.sock_publish.recv()
        assert resp == b'start'

    def listen(self):
        with self.lock_listen:
            self.sock_listen.send(b"listen")
            action, dt, *args = self.sock_listen.recv_multipart()
        if action == b"quit":
            raise self._quit()
        assert action == b"listen"
        dt = self.parse_listen_dt(dt)
        channel = str(args[0], 'ascii')
        data = osgar.lib.serialize.deserialize(args[1])
        return dt, channel, data

    def publish(self, channel, data):
        raw = osgar.lib.serialize.serialize(data)
        with self.lock_publish:
            self.sock_publish.send_multipart([b"publish", bytes(channel, 'ascii'), raw])
            resp, timestamp = self.sock_publish.recv_multipart()
        assert resp == b"publish"
        dt = self.parse_publish_dt(timestamp)
        return dt

    def request_stop(self):
        with self.lock_publish:
            self.sock_publish.send_multipart([b"request_stop"])
            action, dt, *args = self.sock_publish.recv_multipart()

    def is_alive(self):
        with self.lock_publish:
            self.sock_publish.send_multipart([b"is_alive"])
            resp, timestamp = self.sock_publish.recv_multipart()
        assert resp in [b"is_alive", b"quit"], resp
        if resp == b"quit":
            self._quit()
            return False
        return True

    def sleep(self, duration):
        time.sleep(duration)

    def _quit(self):
        g_logger.info(f"{self.name} received quit")
        return SystemExit()


if __name__ == "__main__":
    from .record import main
    main(record)



import unittest
from array import array

from osgar.lib.serialize import serialize, deserialize


class LogUSBTest(unittest.TestCase):

    def test_array(self):
        # USB read returns array('B', [...]), which fails to serialize
        arr = array('B', [1, 2, 3])
        b_arr = serialize(bytes(arr))
        self.assertEqual(deserialize(b_arr), bytes([1, 2, 3]))

# vim: expandtab sw=4 ts=4



"""
  Test basic driving functionality
"""
import math
from datetime import timedelta

from osgar.node import Node
from osgar.lib.mathex import normalizeAnglePIPI
from osgar.bus import BusShutdownException


class Turn(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('desired_speed')
        self.start_pose = None
        self.verbose = False
        self.speed = config['max_speed']
        self.desired_angle = math.radians(config['angle_deg'])
        self.timeout = timedelta(seconds=config['timeout'])
        self.last_position = (0, 0, 0)  # proper should be None, but we really start from zero
        self.max_angular_speed = math.radians(45)
        self.is_moving = None  # unknown

    def send_speed_cmd(self, speed, angular_speed):
        return self.publish('desired_speed', [round(speed*1000), round(math.degrees(angular_speed)*100)])

    def update(self):
        channel = super().update()  # define self.time
        if self.verbose:
            print(self.time, 'Go', channel)
        if channel == 'pose2d':
            x, y, heading = self.pose2d
            pose = (x/1000.0, y/1000.0, math.radians(heading/100.0))
            if self.start_pose is None:
                self.start_pose = pose
            self.last_position = pose
            if self.last_position is not None:
                self.is_moving = (self.last_position != pose)

    def wait(self, dt):  # TODO refactor to some common class
        if self.time is None:
            self.update()
        start_time = self.time
        while self.time - start_time < dt:
            self.update()

    def turn(self, angle, with_stop=True, speed=0.0, timeout=None):
        print(self.time, "turn %.1f" % math.degrees(angle))
        start_pose = self.last_position
        if angle >= 0:
            self.send_speed_cmd(speed, self.max_angular_speed)
        else:
            self.send_speed_cmd(speed, -self.max_angular_speed)
        start_time = self.time
        while abs(normalizeAnglePIPI(start_pose[2] - self.last_position[2])) < abs(angle):
            self.update()
            if timeout is not None and self.time - start_time > timeout:
                print(self.time, "turn - TIMEOUT!")
                break
        if with_stop:
            self.send_speed_cmd(0.0, 0.0)
            start_time = self.time
            while self.time - start_time < timedelta(seconds=2):
                self.update()
                if not self.is_moving:
                    break
            print(self.time, 'stop at', self.time - start_time)

    def run(self):
        try:
            self.send_speed_cmd(0, 0)  # trigger processing (workaround for replay)
            self.update()  # define self.time
            print(self.time, "Turn!")
            self.turn(self.desired_angle, timeout=self.timeout)  # TODO radius
        except BusShutdownException:
            pass


if __name__ == "__main__":
    import argparse
    import os.path
    from osgar.replay import replay
    from osgar.record import record
    from osgar.lib.config import config_load

    parser = argparse.ArgumentParser(description='go')
    subparsers = parser.add_subparsers(help='sub-command help', dest='command')
    subparsers.required = True
    parser_run = subparsers.add_parser('run', help='run on real HW')
    parser_run.add_argument('config', nargs='+', help='configuration file')
    parser_run.add_argument('--note', help='add description')
    parser_run.add_argument('--speed', '-s', help="speed in m/s (default: from config)", type=float)
    parser_run.add_argument('--angle', '-a', help="angle in degrees (default: %(default)sdeg)", type=float, default=90.0)
    parser_run.add_argument('--timeout', help='seconds before stopping (default: %(default)s)', type=int, default=10)

    parser_replay = subparsers.add_parser('replay', help='replay from logfile')
    parser_replay.add_argument('logfile', help='recorded log file')
    parser_replay.add_argument('--force', '-F', dest='force', action='store_true', help='force replay even for failing output asserts')
    parser_replay.add_argument('--config', nargs='+', help='force alternative configuration file')
    parser_replay.add_argument('--verbose', '-v', help="verbose mode", action='store_true')

    args = parser.parse_args()
    print(args)

    if args.command == 'replay':
        args.module = 'app'
        game = replay(args, application=Turn)
        game.verbose = args.verbose
        game.run()

    elif args.command == 'run':
        prefix = 'turn-' + os.path.basename(args.config[0]).split('.')[0] + '-'
        cfg = config_load(*args.config, application=Turn)

        # apply overrides from command line
        cfg['robot']['modules']['app']['init']['angle_deg'] = args.angle
        cfg['robot']['modules']['app']['init']['timeout'] = args.timeout
        if args.speed is not None:
            cfg['robot']['modules']['app']['init']['max_speed'] = args.speed

        record(cfg, prefix)

# vim: expandtab sw=4 ts=4



# dictionary of all available drivers
all_drivers = dict(
    gps="osgar.drivers.gps:GPS",
    imu="osgar.drivers.imu:IMU",
    spider="osgar.drivers.spider:Spider",
    serial="osgar.drivers.logserial:LogSerial",
    can="osgar.drivers.canserial:CANSerial",
    simulator="osgar.drivers.simulator:SpiderSimulator",
    tcp="osgar.drivers.logsocket:LogTCPStaticIP",
    tcpdynamic="osgar.drivers.logsocket:LogTCPDynamicIP",
    tcpserver="osgar.drivers.logsocket:LogTCPServer",
    udp="osgar.drivers.logsocket:LogUDP",
    http="osgar.drivers.logsocket:LogHTTP",
    lidar="osgar.drivers.sicklidar:SICKLidar",
    eduro="osgar.drivers.eduro:Eduro",
    cortexpilot="osgar.drivers.cortexpilot:Cortexpilot",
    usb="osgar.drivers.logusb:LogUSB",
    replay="osgar.drivers.replay:ReplayDriver",
    lordimu="osgar.drivers.lord_imu:LordIMU",
    pcan="osgar.drivers.pcan:PeakCAN",
    kloubak="osgar.drivers.kloubak:RobotKloubak",
    usbcam="osgar.drivers.usbcam:UsbCam",
    rosmsg="osgar.drivers.rosmsg:ROSMsgParser",
    lora="osgar.drivers.lora:LoRa",
    vesc="osgar.drivers.vesc:MotorDriverVESC",
    opencv="osgar.drivers.opencv:LogOpenCVCamera",
    zeromq="osgar.drivers.logzeromq:LogZeroMQ",
    maria="osgar.drivers.maria:RobotMaria",
    timer="osgar.drivers.timer:Timer",
    realsense="osgar.drivers.realsense:RealSense",
    velodyne="osgar.drivers.velodyne:Velodyne",
)


from osgar.lib import create_load_tests
load_tests = create_load_tests(__file__)



#!/usr/bin/env python
"""
  Dummy DARPA scoring server
"""

import os
import sys
import csv
import math
import json
import logging
from collections import defaultdict
from http.server import BaseHTTPRequestHandler, HTTPServer
from mimetypes import guess_type


g_logger = logging.getLogger(__name__)


def dist3d(xyz, xyz2):
    return math.sqrt(sum([(a-b)**2 for a, b in zip(xyz, xyz2)]))


class GameLogic:
    def __init__(self, filename):
        self.score = 0
        self.artf = defaultdict(list)
        with open(filename) as csvfile:
            reader = csv.reader(csvfile)
            for raw in reader:
                if 'artf' in raw:
                    continue
                # artifact name, x, y, z
                artf = raw[0]
                self.artf[artf].append(tuple(float(x) for x in raw[1:]))

    def report_artf(self, artf, xyz):
        if artf in self.artf:
            best = None
            for artf_xyz in self.artf[artf]:
                if best is None or dist3d(xyz, artf_xyz) < best[0]:
                    best = dist3d(xyz, artf_xyz), artf_xyz
            if best is not None and best[0] < 5.0:  # DARPA 5m limit
                self.artf[artf].remove(best[1])
                self.score += 1
                return True
        return False


class MyHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        g_logger.info(f"GET: {self.path}")
        s = self.path.split('/')
        g_logger.info(str(s))
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(b'{"score":%d,"remaining_reports":97,"current_team":"robotika","run_clock":1502.8}' % self.server.game_logic.score)

    def do_POST(self):
        g_logger.info(f"POST: {self.path}")
        s = self.path.split('/')
        assert self.headers['Content-Type'] == 'application/json', self.headers['Content-Type']
        assert 'artifact_reports' in s, s
        size = int(self.headers['Content-Length'])
        data = self.rfile.read(size)
        g_logger.info(f'DATA {data}')
        d = json.loads(data)
        self.server.game_logic.report_artf(d['type'], (d['x'], d['y'], d['z']))
        self.send_response(201)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(b'{"url":"http://10.100.2.200:8000/api/reports/3/","id":3,"x":1.0,"y":2.0,"z":4.0,"type":"Cell Phone","submitted_datetime":"2020-02-18T22:40:05.009145+00:00","run_clock":1505.0,"team":"robotika","run":"0.0.2","report_status":"scored","score_change":1}')


def main():
    import argparse

    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
        datefmt='%Y-%m-%d %H:%M',
    )
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('filename', help='CSV file with header "artf, x, y, z"')
    args = parser.parse_args()
    try:
        server = HTTPServer(('',8888), MyHandler)
        server.game_logic = GameLogic(args.filename)
        print('started httpserver...')
        server.serve_forever()
    except KeyboardInterrupt:
        print('keyboard interrupt')
        server.socket.close()

if __name__ == '__main__':
    main()

# vim: expandtab sw=4 ts=4 




"""
    Control Center for SubT Urban
"""

import threading
import math
import random
import sys
import platform
import json
from datetime import timedelta
from contextlib import suppress

from PyQt5.QtCore import pyqtSignal, QSize, Qt, QPointF, QLineF
from PyQt5.QtGui import QPalette, QKeySequence, QPainter, QColor, QFont, QTransform, QIcon, QPolygonF
from PyQt5.QtWidgets import ( QApplication, QWidget, QPushButton, QVBoxLayout, QHBoxLayout,
                              QMessageBox, QMainWindow, QAction, QToolBar, QMenuBar, QCheckBox,
                              QDockWidget, QComboBox, QMenu)

import osgar.record
import osgar.replay
import osgar.logger
import osgar.bus
import osgar.node
from subt.artifacts import BACKPACK, RESCUE_RANDY, PHONE, VENT, GAS
from subt.report_artf import get_status, report_artf

DEFAULT_ARTF_Z_COORD = 1.0  # when reporting artifacts manually from View

g_filename = None  # filename for replay log

####################################################################################
# dummy functions for testing
def Xget_status():
    return b'{"score":1,"remaining_reports":26,"current_team":"robotika","run_clock":3571.8}'


def Xreport_artf(artf_type, x, y, z):
    return b'{"url":"http://10.100.2.200:8000/api/reports/14/","id":14,"x":68.0,"y":5.0,"z":1.0,"type":"Cell Phone","submitted_datetime":"2020-02-21T19:59:29.520412+00:00","run_clock":3569.5,"team":"robotika","run":"1.5.2","report_status":"scored","score_change":0}'

####################################################################################
g_artf_colors = {
        BACKPACK : Qt.red,
        RESCUE_RANDY : Qt.yellow,
        PHONE : Qt.blue,
        VENT : Qt.white,
        GAS : Qt.darkYellow
    }
ARTF_TYPES = g_artf_colors.keys()

####################################################################################
CFG_LORA = {
  "version": 2,
  "robot": {
    "modules": {
      "cc": {
          "driver": "subt.control_center_qt:OsgarControlCenter",
          "init": {}
      },
      "lora": {
          "driver": "lora",
          "init": {}
      },
      "lora_serial": {
          "driver": "serial",
          "init": {"port": "/dev/lora" if platform.system() == 'Linux' else 'COM35',
                   "speed": 115200}
      },
      "darpa": {
          "driver": "subt.control_center_qt:DARPAReporter",
          "init": {}
      }
    },
    "links": [["lora_serial.raw", "lora.raw"],
              ["lora.raw", "lora_serial.raw"],
              ["lora.robot_status", "cc.robot_status"],
              ["lora.artf", "cc.artf"],
              ["cc.cmd", "lora.cmd"],
              ["cc.artf_xyz", "darpa.artf_xyz"]]
  }
}


class DARPAReporter(osgar.node.Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('get_status', 'report_artf', 'scored')

    def update(self):
        channel = super().update()  # define self.time
        assert channel == 'artf_xyz', channel  # identical to Virtual and LoRa reports

        name, xyz = self.artf_xyz
        x, y, z = [a/1000.0 for a in xyz]

        name2darpa = {RESCUE_RANDY: 'Survivor', BACKPACK: 'Backpack', PHONE: 'Cell Phone',
                      GAS: 'Gas', VENT: 'Vent'}
        artf_type = name2darpa[name]

        before = json.loads(bytes.decode(get_status()))
        self.publish('get_status', before)
        self.sleep(2)
        result = report_artf(artf_type, x, y, z)
        self.publish('report_artf', result)
        self.sleep(2)
        after = json.loads(bytes.decode(get_status()))
        self.publish('get_status', after)
        scored = before['score'] < after['score']
        self.publish('scored', scored)
        self.sleep(2)
        return channel


####################################################################################
class DummyRobot:
    def __init__(self, cfg, bus):
        self.bus = bus
        bus.register('status')
        self.speed = cfg.get('speed', 0.5) # m/s
        self.random = random.Random(cfg.get('seed', 1))
        self.x, self.y = cfg.get('x', 0), cfg.get('y', 0) # m
        self.heading = math.radians(cfg.get('heading', 30)) # degrees
        self.id = cfg.get('id', 1)
        self.thread = threading.Thread(target=self.run)

    def start(self):
        self.thread.start()

    def request_stop(self):
        self.bus.shutdown()

    def join(self, timeout=None):
        self.thread.join(timeout)

    def run(self):
        x, y = self.x, self.y
        heading = self.heading
        last_dt = timedelta()
        current_speed = self.speed
        state = b'Continue'
        try:
            while True:
                dt, channel, data = self.bus.listen()
                #print(dt, channel, data, self.id)
                if channel == 'cmd':
                    robot_id, cmd = data
                    if robot_id == 0 or robot_id == self.id:
                        if cmd == b'Stop':
                            return
                        if cmd == b'GoHome':
                            print(dt, self.id, "GoHome not implemented")
                        state = cmd
                elif channel == 'move':
                    if state == b'Continue':
                        step = (dt - last_dt).total_seconds() * current_speed
                        x += step * math.cos(heading)
                        y += step * math.sin(heading)
                        if self.random.random() > 0.3:
                            heading += math.radians(10)
                    last_dt = dt
                    pose2d = [round(x * 1000), round(y * 1000), round(math.degrees(heading) * 100)]
                    self.bus.publish('status', [self.id, pose2d, state])
        except osgar.bus.BusShutdownException:
            pass


CFG_DEMO = {
  "version": 2,
  "robot": {
    "modules": {
      "cc": {
          "driver": "subt.control_center_qt:OsgarControlCenter",
          "init": {}
      },
      "timer": {
          "driver": "timer",
          "init": {
              "sleep": 0.5,
          }
      },
      "robot1": {
          "driver": "subt.control_center_qt:DummyRobot",
          "init": {
              "id": 1,
              "heading": 0*30,
              "x": 1,
              "y": 0,
          }
      },
      "robot2": {
          "driver": "subt.control_center_qt:DummyRobot",
          "init": {
              "id": 2,
              "heading": 1*30,
              "x": 1,
              "y": 1,
          }
      },
      "robot3": {
          "driver": "subt.control_center_qt:DummyRobot",
          "init": {
              "id": 3,
              "heading": 2*30,
              "x": 0,
              "y": 1,
          }
      },
      "robot4": {
          "driver": "subt.control_center_qt:DummyRobot",
          "init": {
              "id": 4,
              "heading": 3*30,
              "x": -1,
              "y": 1,
          }
      },
      "darpa": {
          "driver": "subt.control_center_qt:DARPAReporter",
          "init": {}
      }
    },
    "links": [
        ["timer.tick", "robot1.move"],
        ["timer.tick", "robot2.move"],
        ["timer.tick", "robot3.move"],
        ["timer.tick", "robot4.move"],
        ["robot1.status", "cc.robot_status"],
        ["robot2.status", "cc.robot_status"],
        ["robot3.status", "cc.robot_status"],
        ["robot4.status", "cc.robot_status"],
        ["cc.cmd", "robot1.cmd"],
        ["cc.cmd", "robot2.cmd"],
        ["cc.cmd", "robot3.cmd"],
        ["cc.cmd", "robot4.cmd"],
        ["cc.artf_xyz", "darpa.artf_xyz"],
    ]
  }
}
####################################################################################


class OsgarControlCenter:

    def __init__(self, init, bus):
        self.bus = bus
        bus.register('cmd', 'artf_xyz')
        self.view = None # will be set from outside by record
        self.thread = None
        self.robot_id = 0  # ALL by default

    def start(self):
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def request_stop(self):
        self.bus.shutdown()

    def join(self, timeout=None):
        self.thread.join(timeout)

    def run(self):
        while True:
            try:
                dt, channel, data = self.bus.listen()
            except osgar.bus.BusShutdownException:
                return
            if channel == "robot_status":
                robot_id, (x, y, heading), status = data
                pose2d = [x/1000, y/1000, math.radians(heading/100)]
                self.view.robot_status.emit(robot_id, pose2d, status)
            elif channel == "artf":
                robot_id, (artf, x, y, z) = data
                print(dt, robot_id, artf, (x/1000, y/1000, z/1000))
                self.view.artf_xyz.emit(artf, (x/1000, y/1000, z/1000))

    def pause_mission(self):
        self.bus.publish('cmd', [self.robot_id, b'Pause'])

    def continue_mission(self):
        self.bus.publish('cmd', [self.robot_id, b'Continue'])

    def stop_mission(self):
        self.bus.publish('cmd', [self.robot_id, b'Stop'])

    def go_home(self):
        self.bus.publish('cmd', [self.robot_id, b'GoHome'])

    def set_robot_id(self, i):
        self.robot_id = i

    def report_artf(self, artf, xyz):
        self.bus.publish('artf_xyz', [artf, [int(a * 1000) for a in xyz]])


class MainWindow(QMainWindow):

    zoom_reset = pyqtSignal()
    zoom_in = pyqtSignal()
    zoom_out = pyqtSignal()

    def __init__(self, arguments):
        super().__init__()
        self.setWindowTitle("Control Center")
        #self.setWindowIcon(QIcon('something.pnp'))
        self._createMenu()
        self._createStatusBar()
        self._createDock()
        self.setCentralWidget(View())
        self.recorder = None
        self.cc = None
        self.centralWidget().show_message.connect(self.status.showMessage)
        self.zoom_reset.connect(self.centralWidget().on_zoom_reset)
        self.zoom_in.connect(self.centralWidget().on_zoom_in)
        self.zoom_out.connect(self.centralWidget().on_zoom_out)
        self.centralWidget().manual_artf.connect(self.on_manual_artf)
        if "--replay" in arguments:
            self.cfg = None
            print(arguments)
            global g_filename
            g_filename = arguments[-1]
        elif "--demo" not in arguments:
            self.cfg = CFG_LORA
            self.status.showMessage("Using LoRa cfg")
        else:
            self.cfg = CFG_DEMO
            self.status.showMessage("Using demo cfg")

    def _createMenu(self):
        self.menu = self.menuBar()
        self.file_menu = self.menu.addMenu("&File")

        exit_action = QAction("Exit", self)
        exit_action.setShortcut(QKeySequence.Quit)
        exit_action.setStatusTip("Exit application")
        exit_action.triggered.connect(self.close)
        exit_action.setIcon(QIcon.fromTheme("application-exit"));
        self.file_menu.addAction(exit_action)

        self.view_menu = self.menu.addMenu("&View")

        zoomin_action = QAction("Zoom in", self)
        # allow multiple shortcuts
        # see https://www.qtcentre.org/threads/1616-Multiple-shortuts-for-the-action
        zoomin_action2 = QAction(zoomin_action)
        zoomin_action.setShortcut(Qt.Key_Plus)
        zoomin_action2.setShortcut(Qt.Key_Equal)
        zoomin_action.triggered.connect(self.zoom_in)
        zoomin_action2.triggered.connect(self.zoom_in)
        self.view_menu.addAction(zoomin_action)
        self.view_menu.addAction(zoomin_action2)

        zoomreset_action = QAction("Zoom reset", self)
        zoomreset_action.setShortcut(Qt.Key_0)
        zoomreset_action.triggered.connect(self.zoom_reset)
        self.view_menu.addAction(zoomreset_action)

        zoomout_action = QAction("Zoom out", self)
        zoomout_action.setShortcut(Qt.Key_Minus)
        zoomout_action.triggered.connect(self.zoom_out)
        self.view_menu.addAction(zoomout_action)

        self.start_stop = QCheckBox("&Recording", self)
        self.start_stop.stateChanged.connect(self.on_start_stop)
        self.menu.setCornerWidget(self.start_stop, Qt.TopRightCorner)

    def _createStatusBar(self):
        self.status = self.statusBar()

    def _createDock(self):
        d = QDockWidget("All Robots", self)
        w = QWidget()
        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignTop)

        self.cb = QComboBox()
        self.cb.addItem("#0 - ALL")
        self.cb.addItem("-----------------")  # 1 is reserved for the Control Center LoRa
        self.cb.addItem("#2 - Eduro")
        self.cb.addItem("#3 - Kloubak K3")
        self.cb.addItem("#4 - MOBoS")
        self.cb.addItem("#5 - Kloubak K2")
        self.cb.addItem("#6 - Maria")
        self.cb.currentIndexChanged.connect(self.selectionchange)
        layout.addWidget(self.cb)

        layout.addWidget(QPushButton("&Pause Mission", clicked=self.on_pause_mission))
        layout.addWidget(QPushButton("&Continue Mission", clicked=self.on_continue_mission))
        layout.addWidget(QPushButton("&Stop Mission", clicked=self.on_stop_mission))
        layout.addWidget(QPushButton("&Go Home", clicked=self.on_go_home))
        w.setLayout(layout)
        d.setWidget(w)
        d.setDisabled(True)
        self.addDockWidget(Qt.RightDockWidgetArea, d)

    def on_start_stop(self, state):
        if self.recorder is None:
            assert state == Qt.Checked
            if g_filename is None:
                self.recorder = record(self.centralWidget(), self.cfg)
            else:
                self.recorder = replay(self.centralWidget(), g_filename)
            self.cc = next(self.recorder)
            print("recording started")
        else:
            assert state == Qt.Unchecked
            with suppress(StopIteration):
                next(self.recorder)
            self.recorder = None
            print("recording stopped")
        for d in self.findChildren(QDockWidget):
            d.setDisabled(self.recorder is None)

    def closeEvent(self, e):
        self.centralWidget().close()

    def on_pause_mission(self):
        print("pause mission", self.cb.currentText())
        self.cc.pause_mission()

    def on_continue_mission(self):
        print('continue mission', self.cb.currentText())
        self.cc.continue_mission()

    def on_stop_mission(self):
        print("stop mission", self.cb.currentText())
        self.cc.stop_mission()

    def on_go_home(self):
        print("go home", self.cb.currentText())
        self.cc.go_home()

    def selectionchange(self, i):
        print("Current index", i, "selection changed ", self.cb.currentText())
        self.cc.set_robot_id(i)

    def on_manual_artf(self, artf, xyz):
        print(artf, xyz)
        self.cc.report_artf(artf, xyz)


def record(view, cfg):
    with osgar.logger.LogWriter(prefix='control-center-', note=str(sys.argv)) as log:
        log.write(0, bytes(str(cfg), 'ascii'))
        recorder = osgar.record.Recorder(config=cfg['robot'], logger=log)
        cc = recorder.modules['cc']
        cc.view = view
        view.robot_statuses = {}
        with recorder:
            yield cc


def replay(view, filename):
    args = type('MyClass', (object,), 
                {'logfile': filename, 'module': 'cc', 'config': None,
                 'debug': None, 'force': True})
    cc = osgar.replay.replay(args)
    cc.view = view
    view.robot_statuses = {}
    cc.start()
    yield cc
    cc.join()


class View(QWidget):

    robot_status = pyqtSignal(int, list, bytes)
    artf_xyz = pyqtSignal(str, tuple)
    show_message = pyqtSignal(str, int)
    manual_artf = pyqtSignal(str, tuple)
    colors = [QColor(Qt.white), QColor(Qt.green), QColor(Qt.blue), QColor(Qt.red),
              QColor(Qt.cyan), QColor(Qt.magenta), QColor(Qt.yellow)]

    def __init__(self):
        super().__init__()
        self.robot_status.connect(self.on_robot_status)
        self.artf_xyz.connect(self.on_artf_xyz)
        self.setAutoFillBackground(True)
        p = self.palette()
        p.setColor(self.backgroundRole(), Qt.black)
        self.setPalette(p)
        self.setMouseTracking(True)
        self.setCursor(Qt.CrossCursor)
        self.translation = QPointF(0, 0) # how many px is the viewport shifted
        self.scale = 100  # 1px == 1cm (position is in meters)
        self.grab_start = None
        self.grab_diff = QPointF(0, 0)
        self.robot_statuses = {}
        self.artifacts = []
        self.reported_artifacts = []

    def on_robot_status(self, robot, pose2d, status):
        #self.show_message.emit(f"robot {robot}: pose2d {pose2d}, status {status}", 3000)
        self.robot_statuses.setdefault(robot, []).append((pose2d, status))
        self.update()

    def on_artf_xyz(self, artf_type, xyz):
        self.artifacts.append((artf_type, xyz))

    def _transform(self):
        t = QTransform() # world transform
        t.translate(self.width() / 2, self.height() / 2) # in pixels
        t.translate(self.translation.x(), self.translation.y())
        t.translate(self.grab_diff.x(), self.grab_diff.y())
        t.scale(self.scale, -self.scale)
        return t

    def paintEvent(self, e):
        transform = self._transform()
        with QPainter(self) as painter:
            self._drawGrid(painter, transform)
            self._drawCenterCross(painter, transform)
            self._drawScale(painter, transform)
            self._drawArtf(painter, transform)
            self._drawReportedArtf(painter, transform)
            for robot, statuses in self.robot_statuses.items():
                base_color = self.colors[robot%len(self.colors)]
                self._drawRobotPath(painter, transform, base_color, statuses)

    def _drawGrid(self, painter, transform):
        painter.setPen(QColor(Qt.white).darker().darker())
        meter = QLineF(transform.map(QPointF(1, 0)), transform.map(QPointF(0, 0))).length()
        limits = [(1.5, 50), (4, 20), (10, 10), (35, 5), (100000, 1)]
        unit = [unit for px,unit in limits if meter < px][0]
        t, ok = transform.inverted()
        r = self.rect()
        x, y = [], []
        for point in [(r.left(),r.top()), (r.left(), r.bottom()), (r.right(), r.top()), (r.right(), r.bottom())]:
            p = t.map(QPointF(*point))
            x.append(p.x())
            y.append(p.y())
        minx, maxx = min(x), max(x)
        miny, maxy = min(y), max(y)
        for i in range(math.ceil(minx/unit), math.ceil(maxx/unit)):
            painter.drawLine(transform.map(QPointF(i * unit, maxy)), transform.map(QPointF(i * unit, miny)))
        for j in range(math.ceil(miny/unit), math.ceil(maxy/unit)):
            painter.drawLine(transform.map(QPointF(maxx, j * unit)), transform.map(QPointF(minx, j * unit)))

    def _drawRobotPath(self, painter, transform, base_color, statuses):
        # just orientation
        t2 = QTransform(math.copysign(1, transform.m11()), transform.m12(),
                        transform.m21(), math.copysign(1, transform.m22()),
                        0, 0)

        painter.setPen(base_color)
        path = QPolygonF()
        for (x, y, heading), status in statuses:
            robot_position = transform.map(QPointF(x, y))
            path.append(robot_position)
            painter.drawEllipse(robot_position, 3, 3)
        # draw direction vector of last position
        facing = QTransform(t2).rotateRadians(heading).map(QPointF(10, 0))
        painter.drawLine(robot_position, robot_position + facing)
        # highlight last position
        painter.setBrush(base_color)
        painter.drawEllipse(robot_position, 4, 4)
        painter.setBrush(Qt.NoBrush)
        # draw path
        painter.setPen(base_color.darker())
        painter.drawPolyline(path)

    def _drawCenterCross(self, painter, transform):
        painter.setPen(Qt.red)
        center = transform.map(QPointF(0, 0))
        painter.drawLine(center - QPointF(10, 0), center + QPointF(10, 0))
        painter.drawLine(center - QPointF(0, 10), center + QPointF(0, 10))

    def _drawScale(self, painter, transform):
        painter.setPen(Qt.red)
        meter = QLineF(transform.map(QPointF(1, 0)), transform.map(QPointF(0, 0))).length()
        painter.drawLine(20, self.height() - 20, 20 + meter, self.height() - 20)
        painter.drawLine(20, self.height() - 25, 20, self.height() - 15)
        painter.drawLine(20 + meter, self.height() - 25, 20 + meter, self.height() - 15)

    def _drawArtf(self, painter, transform):
        for name, xyz in self.artifacts:
            if name in g_artf_colors:
                painter.setPen(g_artf_colors[name])
            else:
                painter.setPen(Qt.gray)
            x, y, z = xyz
            pt = transform.map(QPointF(x, y))
            painter.drawEllipse(pt, 10, 10)

    def _drawReportedArtf(self, painter, transform):
        painter.setPen(Qt.red)
        for name, xyz in self.reported_artifacts:
            x, y, z = xyz
            pt = transform.map(QPointF(x, y))
            radius = transform.map(QPointF(5, 0)).x() - transform.map(QPointF(0, 0)).x()
            painter.setPen(g_artf_colors[name])
            painter.drawEllipse(pt, radius, radius)

    def sizeHint(self):
        return QSize(800, 600)

    def mousePressEvent(self, e):
        if e.buttons() == Qt.LeftButton:
            self.setCursor(Qt.ClosedHandCursor)
            self.grab_start = e.globalPos()
        elif e.buttons() == Qt.RightButton:
            contextMenu = QMenu(self)
            listAct = [contextMenu.addAction(a) for a in ARTF_TYPES]
            action = contextMenu.exec_(self.mapToGlobal(e.pos()))
            if action in listAct:
                t = self._transform()
                t, ok = t.inverted()
                pt = t.map(e.localPos())
                self.reported_artifacts.append((action.text(), (pt.x(), pt.y(), DEFAULT_ARTF_Z_COORD)))
                self.update()
                self.manual_artf.emit(*self.reported_artifacts[-1])

    def mouseReleaseEvent(self, e):
        cursor = self.cursor().shape()
        if cursor == Qt.ClosedHandCursor:
            self.grab_start = None
            self.setCursor(Qt.CrossCursor)
            self.translation += self.grab_diff
            self.grab_diff = QPointF(0, 0)

    def mouseMoveEvent(self, e):
        t = self._transform()
        t, ok = t.inverted()
        p = t.map(e.localPos())

        # TODO: show in tooltip instead?
        self.show_message.emit(f"[{p.x(): .2f}, {p.y(): .2f}]", 0)

        if self.grab_start is not None:
            self.grab_diff = e.globalPos() - self.grab_start
            self.update()

    def on_zoom_reset(self):
        self.scale = 100
        self.update()

    def on_zoom_in(self):
        self.scale *= 1.1
        self.update()

    def on_zoom_out(self):
        self.scale /= 1.1
        self.update()


def main():
    app = QApplication(sys.argv)

    window = MainWindow(app.arguments())
    window.show()
    return app.exec_()


if __name__ == "__main__":
    sys.exit(main())

# vim: expandtab sw=4 ts=4



"""
  Wrapper for CAN-serial communication and control of CAN bridge
"""

import struct
from threading import Thread
from collections import OrderedDict

from osgar.bus import BusShutdownException


# The CAN message format is specific to CAN bridge used. The first two bytes ("header")
# contain CAN-ID (11bits = 4bits function code and 7bits node ID), RTR (1bit) and the packet
# length (4bits). The bridges uses 0xFF and 0xFE for internal communication control.


CAN_BRIDGE_READY = b'\xfe\x10'  # CAN bridge is ready to accept configuration commands
CAN_BRIDGE_SYNC = b'\xFF'*10    # CAN bridge synchronization bytes
CAN_BRIDGE_START = b'\xfe\x31'  # start bridge

CAN_SPEED = OrderedDict([
                ('10k', b'\xfe\x50'),
                ('20k', b'\xfe\x51'),
                ('50k', b'\xfe\x52'),
                ('125k', b'\xfe\x53'),
                ('250k', b'\xfe\x54'),
                ('500k', b'\xfe\x55'),
                ('800k', b'\xfe\x56'),
                ('1M', b'\xfe\x57')
            ])

# CANopen network management (NMT) communication
# https://en.wikipedia.org/wiki/CANopen#Network_management_(NMT)_protocols
NMT_STATE_BOOT_UP = 0
NMT_STATE_STOPPED = 4
NMT_STATE_OPERATIONAL = 5
NMT_STATE_PREOPERATIONAL = 0x7F

NMT_CMD_GO_TO_OPERATIONAL = 0x01
NMT_CMD_GO_TO_STOPPED = 0x02
NMT_CMD_GO_TO_PREOPERATIONAL = 0x80
NMT_CMD_GO_TO_RESET_NODE = 0x81
NMT_CMD_GO_TO_RESET_COMMUNICATION = 0x82


def is_nmt_message(msg_id):
    # NMT node monitoring (node guarding/heartbeat)
    # 700 + NodeID
    assert (msg_id & 0xF800) == 0, hex(msg_id)  # is not msg_id used for other CAN bridge features?!
    return (msg_id & 0x780) == 0x700  # 4 bit function code, 11 bit node ID


def get_node_id(msg_id):
    return msg_id & 0x7F  # 7 bits address in CANopen


def CAN_packet(msg_id, data):
    header = [(msg_id>>3) & 0xff, (msg_id<<5) & 0xe0 | (len(data) & 0xf)]
    if isinstance(data, bytes):
        return bytes(header) + data
    return bytes(header + data)


def parse_header(data):
    assert len(data) >= 2, len(data)
    msg_id = (data[0] << 3) | (data[1] >> 5)
    rtr = (data[1] >> 4) & 0x1  # Remote transmission request
    size = data[1] & 0x0f
    return msg_id, rtr, size


def print_packet(data, dbc = {}):
    assert len(data) >= 2, len(data)
    msg_id, rtr, size = parse_header(data)
    if rtr:
        return [hex(x) for x in data[:2]]
    else:
        assert len(data) == 2 + size, (len(data), 2 + size)
        if msg_id in dbc:
            return hex(msg_id), [x for x in struct.unpack(dbc[msg_id], data[2:])]
        else:
            return hex(msg_id), [hex(x) for x in data[2:]]


class CANSerial(Thread):
    def __init__(self, config, bus):
        bus.register('can', 'raw')
        Thread.__init__(self)
        self.setDaemon(True)

        self.bus = bus
        self.buf = b''

        speed = config.get('speed', '1M')  # default 1Mbit
        if speed not in CAN_SPEED:
            raise ValueError('unsupported speed: {}\nUse:{}'.format(speed, list(CAN_SPEED.keys())))
        self.can_speed_cmd = CAN_SPEED[speed]
        self.is_canopen = config.get('canopen', False)
        self.can_bridge_initialized = False
        self.modules_for_restart = set()

    @staticmethod
    def split_buffer(data):
        # skip 0xFF prefix bytes (CAN bridge control bytes)
        data = data.lstrip(b'\xff')

        if len(data) >= 2:
            # see https://en.wikipedia.org/wiki/CAN_bus
            __, rtr, size = parse_header(data)
            if rtr:
                return data[2:], data[:2]
            elif len(data) >= 2 + size:
                return data[2+size:], data[:2+size]
        return data, b''  # no complete packet available yet

    def send_data(self, module_id, data):
        self.bus.publish('raw', CAN_packet(module_id, data))

    def read_packet(self):
        while True:
            self.buf, packet = self.split_buffer(self.buf)
            if len(packet) > 0:
                msg_id, __, __ = parse_header(packet)
                return msg_id, packet[2:]
            else:
                dt, channel, data = self.bus.listen()
                if channel == 'raw':
                    self.buf += data
                else:
                    print('Ignoring', channel)

    def reset_all_modules(self):
        """
           Reset all modules:
           - broadcast "reset node"
           - make sure every node reaches pre-operational state
           - switch all nodes to operational state
           
           return: list of operational modules
        """
        # reset all nodes
        # addr = 0 means, all nodes must listen
        # data[1] = 0 means, all nodes should act
        self.send_data(0, [NMT_CMD_GO_TO_RESET_NODE, 0])

        ackBootup = [] # Heart Beat 0, bootup, after reset
        ackPreop = [] # HB 127
        ackOp = [] # HB 5

        # The number of modules is unknown, but there is at least one expected.
        # All modules should go through NMT_STATE_BOOT_UP and then 
        # NMT_STATE_PREOPERATIONAL states.
        # If some node fails to switch to preoperation then this loop
        # is going to be infinite! In normal operation finish when
        # all nodes pass through both states, i.e. ==
        while len(ackBootup) == 0 or len(ackBootup) > len(ackPreop):
            msg_id, data = self.read_packet()
            if is_nmt_message(msg_id):
                nodeID = get_node_id(msg_id)
                state = data[0]

                if state == NMT_STATE_BOOT_UP:
                    print("Started module", nodeID)
                    ackBootup.append(nodeID)

                if state == NMT_STATE_PREOPERATIONAL:
                    if nodeID in ackBootup:
                        print("Module", nodeID, "in preoperation.")
                        ackPreop.append(nodeID)
                    else:
                        print("WARNING!!! Module", nodeID, "preop BEFORE bootup!!!")

        print("------- Switch to Operation mode --------")
        self.send_data(0, [NMT_CMD_GO_TO_OPERATIONAL, 0])  # operation mode
        # The same comment as above: if some module refuses to start this loop will
        # be infinite. Otherwise it collects data until all detected modules
        # switch to operational.
        while len(ackPreop) > len(ackOp):
            msg_id, data = self.read_packet()
            if is_nmt_message(msg_id):
                nodeID = get_node_id(msg_id)
                state = data[0]

                if state == NMT_STATE_OPERATIONAL:
                    if nodeID in ackPreop:
                        print("Module", nodeID, "in operation.")
                        ackOp.append(nodeID)
                    else:
                        print("WARNING!!! Module", nodeID, "op BEFORE preop!!!")

        print("collecting some packets ...")
        countHB = 0  # count heart beats
        while countHB < len(ackOp) * 3: # ie approx 3s
            msg_id, data = self.read_packet()
            if is_nmt_message(msg_id):
                countHB += 1
                if countHB % len(ackOp) == 0:
                    print(countHB / len(ackOp), '...')
                assert len(data) == 1, len(data)
                state = data[0]
                if state != NMT_STATE_OPERATIONAL:
                    print('ERROR - module', get_node_id(msg_id), 'data', data)
        return ackOp

    def check_and_restart_module(self, module_id, status):
        if status != NMT_STATE_OPERATIONAL:
            if module_id not in self.modules_for_restart:
                print("RESET", module_id)
                self.bus.publish('raw', CAN_packet(0, [NMT_CMD_GO_TO_RESET_NODE, module_id]))
                self.modules_for_restart.add(module_id)

            elif status == NMT_STATE_PREOPERATIONAL:  # restarted and in preoperation
                print("SWITCH TO OPERATION", module_id)
                self.bus.publish('raw', CAN_packet(0, [NMT_CMD_GO_TO_OPERATIONAL, module_id]))

        elif module_id in self.modules_for_restart:
            print("RUNNING", module_id)
            self.modules_for_restart.remove(module_id)

    def process_packet(self, packet):
        if packet == CAN_BRIDGE_READY:
            self.bus.publish('raw', CAN_BRIDGE_SYNC)
            self.bus.publish('raw', self.can_speed_cmd)
            self.bus.publish('raw', CAN_BRIDGE_START)
            self.can_bridge_initialized = True
            if self.is_canopen:
# HACK!                self.reset_all_modules()  # TODO config
                self.bus.publish('raw', CAN_packet(0, [NMT_CMD_GO_TO_OPERATIONAL, 0]))
            return None

        msg_id, __, __ = parse_header(packet)
        if is_nmt_message(msg_id):
            assert len(packet) == 3, len(packet)
            self.check_and_restart_module(get_node_id(msg_id), packet[2])
        return packet

    def process_gen(self, data):
        self.buf, packet = self.split_buffer(self.buf + data)
        while len(packet) > 0:
            ret = self.process_packet(packet)
            if ret is not None:
                yield ret
            self.buf, packet = self.split_buffer(self.buf)  # i.e. process only existing buffer now

    def slot_raw(self, timestamp, data):
        if len(data) > 0:
            for packet in self.process_gen(data):
                msg_id, rtr, size = parse_header(packet)
                if rtr == 0:
                    assert size + 2 == len(packet), (size, len(packet))
                # TODO verify how rtr is handled on PCAN?
                self.bus.publish('can', [msg_id, packet[2:], 0])

    def slot_can(self, timestamp, data):
        if self.can_bridge_initialized:
            # at the moment is can serial just forwarding translation of raw packets
            msg_id, payload, flags = data
            assert flags == 0, flags  # CAN bridge supports only basic addressing
            self.bus.publish('raw', CAN_packet(msg_id, payload))
        else:
            print('CAN bridge not initialized yet!')

    def run(self):
        try:
            while True:
                dt, channel, data = self.bus.listen()
                if channel == 'raw':
                    self.slot_raw(dt, data)
                elif channel == 'can':
                    self.slot_can(dt, data)
                else:
                    assert False, channel  # unsupported input channel
        except BusShutdownException:
            if self.is_canopen:
                self.bus.publish('raw', CAN_packet(0, [NMT_CMD_GO_TO_PREOPERATIONAL, 0]))  # pre-operational

    def request_stop(self):
        self.bus.shutdown()


if __name__ == "__main__":
    import argparse
    from osgar.logger import LogReader, lookup_stream_id
    from osgar.lib.serialize import deserialize

    parser = argparse.ArgumentParser(description='Parse CAN stream messages')
    parser.add_argument('logfile', help='filename of stored file')
    parser.add_argument('--stream', help='stream ID or name', default='can.can')
    parser.add_argument('--dbc', help='interpretation of raw data',
                        choices=['spider', 'eduro', 'kloubak'])
    args = parser.parse_args()

    dbc = {}
    if args.dbc == 'spider':
        dbc = {
            0x200: 'H',     # status
            0x201: 'HHHH',  # wheels
            0x202: 'HHHH',  # drive status
            0x203: 'HHHH',  # zero steering
            0x204: 'HHBBH'  # user input
        }
    elif args.dbc == 'eduro':
        dbc = {
#            0x80:  # SYNC, no data
            0x181: '<i',    # encoders L
            0x182: '<i',    # encoders R
            #0x187, 0x387, 0x487  # compass, acc
            0x28A: '<H',    # buttons
            0x18B: '<H',    # battery(V)/100.0
        }
    elif args.dbc == 'kloubak':
        dbc = {
            0x101: '>i',     # cmd1
            0x102: '>i',     # cmd2
            0x91: '>ihh',  # encoders
            0x92: '>ihh',
            0x93: '>ihh',
            0x94: '>ihh',
        }

    stream_id = lookup_stream_id(args.logfile, args.stream)
    with LogReader(args.logfile, only_stream_id=stream_id) as log:
        for timestamp, stream_id, data in log:
            print(timestamp, print_packet(deserialize(data), dbc))

# vim: expandtab sw=4 ts=4



"""
  pyrealsense2 OSGAR wrapper

"""
import math
import logging
import threading
import numpy as np

g_logger = logging.getLogger(__name__)

try:
    import pyrealsense2 as rs
    try:
        prefix = rs.pyrealsense2.__version__[:4]
        if prefix in ('2.32', '2.33'):
            g_logger.error(f"RealSense version {rs.pyrealsense2.__version__} is broken in several ways!")
    except Exception as e:
        g_logger.error('pyrealsense2 version not found due to: ' + str(e))
except:
    g_logger.info('RealSense not installed!')
    from unittest.mock import MagicMock
    rs = MagicMock()

from osgar.node import Node
from osgar.lib import quaternion


# https://github.com/IntelRealSense/librealsense/blob/master/doc/t265.md#sensor-origin-and-coordinate-system

def t265_to_osgar_position(t265_position):
    x = -t265_position.z
    y = -t265_position.x
    z = t265_position.y
    return [x, y, z]


def t265_to_osgar_orientation(t265_orientation):
    x0 = t265_orientation.z
    y0 = -t265_orientation.x
    z0 = t265_orientation.y
    w0 = t265_orientation.w
    return [x0, y0, z0, w0]


class RealSense(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('pose2d', 'pose3d', 'pose_raw', 'orientation', 'depth:gz', 'color')
        self.verbose = config.get('verbose', False)
        self.depth_subsample = config.get("depth_subsample", 3)
        self.pose_subsample = config.get("pose_subsample", 20)
        self.depth_rgb = config.get("depth_rgb", False)
        self.pose_pipeline = None  # not initialized yet
        self.depth_pipeline = None
        self.finished = None
        if self.depth_rgb:
            import cv2
            global cv2

    def pose_callback(self, frame):
        try:
            pose = frame.as_pose_frame().get_pose_data()
            n = frame.get_frame_number()
            if n % self.pose_subsample != 0:
                return
            timestamp = frame.get_timestamp()
        except Exception as e:
            print(e)
            self.finished.set()
            return
        orientation = t265_to_osgar_orientation(pose.rotation)
        self.publish('orientation', orientation)
        x, y, z = t265_to_osgar_position(pose.translation)
        yaw = quaternion.heading(orientation)
        self.publish('pose2d', [int(x * 1000), int(y * 1000), int(math.degrees(yaw) * 100)])
        self.publish('pose3d', [[x, y, z], orientation])
        self.publish('pose_raw', [n, timestamp,
                             [pose.translation.x, pose.translation.y, pose.translation.z],
                             [pose.rotation.x, pose.rotation.y, pose.rotation.z, pose.rotation.w],
                             [pose.velocity.x, pose.velocity.y, pose.velocity.z],
                             [pose.angular_velocity.x, pose.angular_velocity.y, pose.angular_velocity.z],
                             [pose.acceleration.x, pose.acceleration.y, pose.acceleration.z],
                             [pose.angular_acceleration.x, pose.angular_acceleration.y,
                              pose.angular_acceleration.z],
                             [pose.mapper_confidence, pose.tracker_confidence],
                             ])  # raw RealSense2 Pose data

    def depth_callback(self, frameset):
        try:
            assert frameset.is_frameset()
            depth_frame = frameset.as_frameset().get_depth_frame()
            n = depth_frame.get_frame_number()
            if n % self.depth_subsample != 0:
                return
            assert depth_frame.is_depth_frame()
            depth_image = np.asanyarray(depth_frame.as_depth_frame().get_data())

            if self.depth_rgb:
                color_frame = frameset.as_frameset().get_color_frame()
                assert color_frame.is_video_frame()
                color_image = np.asanyarray(color_frame.as_video_frame().get_data())
                __, data = cv2.imencode('*.jpeg', color_image)

        except Exception as e:
            print(e)
            self.finished.set()
            return

        self.publish('depth', depth_image)
        if self.depth_rgb:
            self.publish('color', data.tobytes())


    def start(self):
        self.finished = threading.Event()
        ctx = rs.context()
        device_list = ctx.query_devices()
        if len(device_list) == 0:
            g_logger.warning("No RealSense devices detected!")
            self.finished.set()
            return

        enable_pose, enable_depth = False, False
        for device in device_list:
            name = device.get_info(rs.camera_info.name)
            serial_number = device.get_info(rs.camera_info.serial_number)
            intro = f"Found {name} (S/N: {serial_number}): "
            product_line = device.get_info(rs.camera_info.product_line)
            if product_line == "D400":
                info_msg = "Enabling streams: depth"
                enable_depth = True
                if self.depth_rgb:
                    info_msg += ", depth_rgb"
                g_logger.info(intro + info_msg)
            elif product_line == "T200":
                enable_pose = True
                g_logger.info(intro + "Enabling pose stream")
            else:
                g_logger.warning(f"Unknown type: {product_line}")

        if enable_pose:
            self.pose_pipeline = rs.pipeline(ctx)
            pose_cfg = rs.config()
            pose_cfg.enable_stream(rs.stream.pose)
            self.pose_pipeline.start(pose_cfg, self.pose_callback)

        if enable_depth:
            self.depth_pipeline = rs.pipeline(ctx)
            depth_cfg = rs.config()
            depth_cfg.enable_stream(rs.stream.depth, 640, 360, rs.format.z16, 30)
            if self.depth_rgb:
                depth_cfg.enable_stream(rs.stream.color, 960, 540, rs.format.bgr8, 30)
            self.depth_pipeline.start(depth_cfg, self.depth_callback)

        if not enable_pose and not enable_depth:
            self.finished.set()

    def request_stop(self):
        if self.pose_pipeline is not None:
            self.pose_pipeline.stop()
        if self.depth_pipeline is not None:
            self.depth_pipeline.stop()
        super().request_stop()
        self.finished.set()

    def join(self, timeout_sec=None):
        self.finished.wait(timeout_sec)


# vim: expandtab sw=4 ts=4



import unittest
import datetime
from unittest.mock import MagicMock

from osgar.drivers.winsen_gas_detector import WinsenCO2


class WinsenCO2Test(unittest.TestCase):

    def test_parse_packet(self):
        ref_packet = bytes.fromhex('ff86063b3d000000fc')
        sensor = WinsenCO2(config={}, bus=MagicMock())
        sensor._buf += ref_packet
        packet = sensor.get_packet()
        self.assertEqual(packet, ref_packet)

        co2_value = sensor.parse_CO2_packet(packet)
        self.assertEqual(co2_value, 1595)

    def test_clear_buffer(self):
        sensor = WinsenCO2(config={}, bus=MagicMock())
        sensor._buf = bytes([0] * 1000)
        packet = sensor.get_packet()
        self.assertIsNone(packet)
        self.assertEqual(sensor._buf, b'')

    def test_keep_ff_in_buffer(self):
        sensor = WinsenCO2(config={}, bus=MagicMock())
        sensor._buf = bytes([0] * 1000) + b'\xFF'
        packet = sensor.get_packet()
        self.assertIsNone(packet)
        self.assertEqual(sensor._buf, bytes([0xFF]))

    def test_checksum_error(self):
        sensor = WinsenCO2(config={}, bus=MagicMock())
        sensor._buf = bytes.fromhex('ff86063b3d000000fe')  # valid should be 0xFC
        packet = sensor.get_packet()
        self.assertIsNone(packet)

        self.assertEqual(sensor.errors, 1)

# vim: expandtab sw=4 ts=4



"""
  Extra math routines
"""
import math


def normalizeAnglePIPI( angle ):
    while angle < -math.pi:
        angle += 2*math.pi
    while angle > math.pi:
        angle -= 2*math.pi
    return angle 

# vim: expandtab sw=4 ts=4



"""
  DARPA E-stop Tier level 2
"""
from osgar.node import Node
from osgar.bus import BusShutdownException

ATIS_FRAME_PACKET = bytes([0x7E, 0x00, 0x04, 0x08, 0x01, 0x49, 0x53, 0x5A])
NORMAL_OPERATION_PACKET = bytes([0x7E, 0x00, 0x0B, 0x88, 0x01, 0x49, 0x53, 0x00,
                                 0x01, 0x00, 0x02, 0x00, 0x00, 0x00, 0xD7])
EMERGENCY_STOP_PACKET = bytes([0x7E, 0x00, 0x0B, 0x88, 0x01, 0x49, 0x53, 0x00,
                               0x01, 0x00, 0x02, 0x00, 0x00, 0x02, 0xD5])
HEADER = bytes([0x7E, 0x00, 0x0B, 0x88, 0x01, 0x49, 0x53])
PACKET_SIZE = 15  # including header and checksum

MASTER_STOP = bytes.fromhex('7E 00 10 17 01 00 00 00 00 00 00 FF FF FF FE 03 44 31 05 6F')


class EStop(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('emergency_stop', 'raw')
        self._buf = b''
        self.master = config.get('master', False)

    def update(self):  # hack, this method should be called run instead!
        channel = super().update()  # define self.time
        assert channel == "raw", channel
        self._buf += self.raw

        packet = None
        while HEADER in self._buf and self._buf.index(HEADER) + PACKET_SIZE <= len(self._buf):
            i = self._buf.index(HEADER)
            packet = self._buf[i:i+PACKET_SIZE]
            self._buf = self._buf[i+PACKET_SIZE:]       

        if packet is not None:
            print('packet:', len(packet), packet.hex())
            if packet[-2] == 0x02:
                self.publish('emergency_stop', True)
                self.request_stop()
            self.sleep(1.0)
            if self.master:
                self.publish('raw', MASTER_STOP)
            else:
                self.publish('raw', ATIS_FRAME_PACKET)

    def run(self):
        try:
            if self.master:
                self.publish('raw', MASTER_STOP)
            else:
                self.publish('raw', ATIS_FRAME_PACKET)
            while True:
                self.update()
        except BusShutdownException:
            pass

# vim: expandtab sw=4 ts=4
